# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""

import datetime
import warnings
from collections import namedtuple
from typing import Any

import pytest
import yaml
from antimeridian import FixWindingWarning
from dateutil.tz import UTC

import datacube.scripts.cli_app
import datacube.scripts.search_tool
from datacube import Datacube
from datacube.cfg import ODCEnvironment
from datacube.cfg.opt import _DEFAULT_DB_USER
from datacube.index import Index
from datacube.model import Dataset, Range
from datacube.testutils import suppress_deprecations
from datacube.utils.dates import tz_as_utc

from .search_utils import _cli_csv_search, _csv_search_raw, _load_product_query


def test_search_by_metadata(index: Index, ls8_eo3_product, wo_eo3_product) -> None:
    lds = list(
        index.products.search_by_metadata({"properties": {"product_family": "ard"}})
    )
    assert len(lds) == 0
    lds = list(
        index.products.search_by_metadata({"properties": {"odc:product_family": "ard"}})
    )
    assert len(lds) == 1
    lds = list(
        index.products.search_by_metadata({"properties": {"platform": "landsat-8"}})
    )
    assert len(lds) == 0
    lds = list(
        index.products.search_by_metadata({"properties": {"eo:platform": "landsat-8"}})
    )
    assert len(lds) == 1


def test_find_most_recent_change(
    index: Index, ls8_eo3_dataset, ls8_eo3_dataset2, ls8_eo3_dataset3
) -> None:
    product = ls8_eo3_dataset.product
    dt = index.products.most_recent_change(product)
    assert dt == ls8_eo3_dataset3.indexed_time
    index.datasets.archive([ls8_eo3_dataset.id, ls8_eo3_dataset2.id])
    dt = index.products.most_recent_change(product.name)
    d = index.datasets.get(ls8_eo3_dataset2.id)
    assert d is not None
    assert dt == d.archived_time


def test_search_dataset_equals_eo3(index: Index, ls8_eo3_dataset: Dataset) -> None:
    datasets = list(index.datasets.search(platform="landsat-8"))
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    datasets = list(index.datasets.search(platform="landsat-8", instrument="OLI_TIRS"))
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # Wrong product family
    with pytest.raises(ValueError):
        next(
            index.datasets.search(
                platform="landsat-8",
                product_family="splunge",
            )
        )


def test_search_dataset_range_eo3(
    index: Index,
    ls8_eo3_dataset: Dataset,
    ls8_eo3_dataset2: Dataset,
    ls8_eo3_dataset3: Dataset,
    ls8_eo3_dataset4: Dataset,
) -> None:
    # Less Than
    datasets = list(
        index.datasets.search(
            product=ls8_eo3_dataset.product.name, cloud_cover=Range(None, 50.0)
        )
    )
    assert len(datasets) == 2
    ids = [ds.id for ds in datasets]
    assert ls8_eo3_dataset3.id in ids
    assert ls8_eo3_dataset4.id in ids

    # Greater than
    datasets = list(
        index.datasets.search(
            product=ls8_eo3_dataset.product.name, cloud_cover=Range(50.0, None)
        )
    )
    assert len(datasets) == 2
    ids = [ds.id for ds in datasets]
    assert ls8_eo3_dataset.id in ids
    assert ls8_eo3_dataset2.id in ids

    # Full Range comparison
    datasets = list(
        index.datasets.search(
            product=ls8_eo3_dataset.product.name, cloud_cover=Range(20.0, 55.0)
        )
    )
    assert len(datasets) == 2
    ids = [ds.id for ds in datasets]
    assert ls8_eo3_dataset2.id in ids
    assert ls8_eo3_dataset3.id in ids


def test_search_dataset_by_metadata_eo3(index: Index, ls8_eo3_dataset: Dataset) -> None:
    datasets = index.datasets.search_by_metadata(
        {"properties": {"eo:platform": "landsat-8", "eo:instrument": "OLI_TIRS"}}
    )
    datasets = list(datasets)
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    datasets = index.datasets.search_by_metadata(
        {"properties": {"eo:platform": "landsat-5", "eo:instrument": "TM"}}
    )
    datasets = list(datasets)
    assert len(datasets) == 0
    datasets = index.datasets.search_by_metadata(
        {"properties": {"eo:platform": "landsat-8", "eo:instrument": "OLI_TIRS"}},
        archived=None,
    )
    datasets = list(datasets)
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id
    datasets = index.datasets.search_by_metadata(
        {"properties": {"eo:platform": "landsat-8", "eo:instrument": "OLI_TIRS"}},
        archived=True,
    )
    datasets = list(datasets)
    assert len(datasets) == 0

    index.datasets.archive([ls8_eo3_dataset.id])

    datasets = index.datasets.search_by_metadata(
        {"properties": {"eo:platform": "landsat-8", "eo:instrument": "OLI_TIRS"}},
        archived=None,
    )
    datasets = list(datasets)
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id
    datasets = index.datasets.search_by_metadata(
        {"properties": {"eo:platform": "landsat-8", "eo:instrument": "OLI_TIRS"}},
        archived=True,
    )
    datasets = list(datasets)
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id


def test_search_day_eo3(index: Index, ls8_eo3_dataset: Dataset) -> None:
    # Matches day
    datasets = list(index.datasets.search(time=datetime.date(2016, 5, 12)))
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # Different day: no match
    datasets = list(index.datasets.search(time=datetime.date(2016, 5, 13)))
    assert len(datasets) == 0


def test_search_dataset_ranges_eo3(index: Index, ls8_eo3_dataset: Dataset) -> None:
    # In the lat bounds.
    datasets = list(
        index.datasets.search(
            lat=Range(-37.5, -36.5),
            time=Range(
                datetime.datetime(2016, 5, 12, 23, 0, 0),
                datetime.datetime(2016, 5, 12, 23, 59, 59),
            ),
        )
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # Out of the lat bounds.
    datasets = list(
        index.datasets.search(
            lat=Range(28, 32),
            time=Range(
                datetime.datetime(2016, 5, 12, 23, 0, 0),
                datetime.datetime(2016, 5, 12, 23, 59, 59),
            ),
        )
    )
    assert len(datasets) == 0

    # Out of the time bounds
    datasets = list(
        index.datasets.search(
            lat=Range(-37.5, -36.5),
            time=Range(
                datetime.datetime(2014, 7, 26, 21, 48, 0),
                datetime.datetime(2014, 7, 26, 21, 50, 0),
            ),
        )
    )
    assert len(datasets) == 0

    # A dataset that overlaps but is not fully contained by the search bounds.
    # Should we distinguish between 'contains' and 'overlaps'?
    datasets = list(index.datasets.search(lat=Range(-40, -37.1)))
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # Single point search
    datasets = list(
        index.datasets.search(
            lat=-37.0,
            time=Range(
                datetime.datetime(2016, 5, 12, 23, 0, 0),
                datetime.datetime(2016, 5, 12, 23, 59, 59),
            ),
        )
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    datasets = list(
        index.datasets.search(
            lat=30.0,
            time=Range(
                datetime.datetime(2016, 5, 12, 23, 0, 0),
                datetime.datetime(2016, 5, 12, 23, 59, 59),
            ),
        )
    )
    assert len(datasets) == 0

    # Single timestamp search
    datasets = list(
        index.datasets.search(
            lat=Range(-37.5, -36.5),
            time=datetime.datetime(2016, 5, 12, 23, 50, 40),
        )
    )
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    datasets = list(
        index.datasets.search(
            lat=Range(-37.5, -36.5), time=datetime.datetime(2016, 5, 12, 23, 0, 0)
        )
    )
    assert len(datasets) == 0


def test_zero_width_range_search(index: Index, ls8_eo3_dataset4: Dataset) -> None:
    # Test time search against zero-width time metadata
    datasets = list(
        index.datasets.search(
            time=Range(
                begin=datetime.datetime(
                    2013, 7, 21, 0, 57, 26, 432563, tzinfo=datetime.timezone.utc
                ),
                end=datetime.datetime(
                    2013, 7, 21, 0, 57, 26, 432563, tzinfo=datetime.timezone.utc
                ),
            )
        )
    )
    assert len(datasets) == 1

    datasets = list(
        index.datasets.search(
            time=Range(
                begin=datetime.datetime(
                    2013, 7, 21, 0, 57, 26, 432563, tzinfo=datetime.timezone.utc
                ),
                end=datetime.datetime(
                    2013, 7, 21, 0, 57, 27, 432563, tzinfo=datetime.timezone.utc
                ),
            )
        )
    )
    assert len(datasets) == 1

    datasets = list(
        index.datasets.search(
            time=Range(
                begin=datetime.datetime(
                    2013, 7, 21, 0, 57, 25, 432563, tzinfo=datetime.timezone.utc
                ),
                end=datetime.datetime(
                    2013, 7, 21, 0, 57, 26, 432563, tzinfo=datetime.timezone.utc
                ),
            )
        )
    )
    assert len(datasets) == 1


def test_search_globally_eo3(index: Index, ls8_eo3_dataset: Dataset) -> None:
    # No expressions means get all.
    results = list(index.datasets.search())
    assert len(results) == 1

    # Dataset sources aren't loaded by default
    assert results[0].sources is None


def test_search_by_product_eo3(
    index: Index,
    base_eo3_product_doc: dict,
    ls8_eo3_dataset: Dataset,
    wo_eo3_dataset: Dataset,
) -> None:
    # Query all the test data, the counts should match expected
    results = _load_product_query(index.datasets.search_by_product())
    assert len(results) == 2
    dataset_count = sum(len(ds) for ds in results.values())
    assert dataset_count == 2

    # Query one product
    products = _load_product_query(
        index.datasets.search_by_product(platform="landsat-8", product_family="wo")
    )
    assert len(products) == 1
    [dataset] = products[base_eo3_product_doc["name"]]
    assert dataset.id == wo_eo3_dataset.id
    assert dataset.is_eo3
    with suppress_deprecations():
        assert dataset.type == dataset.product  # DEPRECATED MEMBER


def test_search_limit_eo3(
    index: Index, ls8_eo3_dataset, ls8_eo3_dataset2, wo_eo3_dataset
) -> None:
    prod = ls8_eo3_dataset.product.name
    datasets = list(index.datasets.search(product=prod))
    assert len(datasets) == 2
    datasets = list(index.datasets.search(limit=1, product=prod))
    ids = [ds.id for ds in datasets]
    assert len(ids) == 1
    assert len(datasets) == 1
    datasets = list(index.datasets.search(limit=0, product=prod))
    assert len(datasets) == 0
    datasets = list(index.datasets.search(limit=5, product=prod))
    assert len(datasets) == 2

    datasets = list(index.datasets.search_returning(("id",), product=prod))
    assert len(datasets) == 2
    datasets = list(index.datasets.search_returning(("id",), limit=1, product=prod))
    assert len(datasets) == 1
    datasets = list(index.datasets.search_returning(("id",), limit=0, product=prod))
    assert len(datasets) == 0
    datasets = list(index.datasets.search_returning(("id",), limit=5, product=prod))
    assert len(datasets) == 2

    # Limit is per product not overall.  (But why?!?)
    datasets = list(index.datasets.search())
    assert len(datasets) == 3
    datasets = list(index.datasets.search(limit=1))
    assert len(datasets) == 2
    datasets = list(index.datasets.search(limit=0))
    assert len(datasets) == 0
    datasets = list(index.datasets.search(limit=5))
    assert len(datasets) == 3

    datasets = list(index.datasets.search_returning(("id",)))
    assert len(datasets) == 3
    datasets = list(index.datasets.search_returning(("id",), limit=1))
    assert len(datasets) == 2
    datasets = list(index.datasets.search_returning(("id",), limit=0))
    assert len(datasets) == 0
    datasets = list(index.datasets.search_returning(("id",), limit=5))
    assert len(datasets) == 3


def test_search_archived_eo3(
    index: Index, ls8_eo3_dataset, ls8_eo3_dataset2, wo_eo3_dataset
) -> None:
    prod = ls8_eo3_dataset.product.name
    datasets = list(index.datasets.search(archived=False, product=prod))
    assert len(datasets) == 2
    datasets = list(index.datasets.search(archived=None, product=prod))
    assert len(datasets) == 2
    datasets = list(index.datasets.search(archived=True, product=prod))
    assert len(datasets) == 0

    index.datasets.archive([ls8_eo3_dataset.id])

    datasets = list(index.datasets.search(archived=False, product=prod))
    assert len(datasets) == 1
    datasets = list(index.datasets.search(archived=None, product=prod))
    assert len(datasets) == 2
    datasets = list(index.datasets.search(archived=True, product=prod))
    assert len(datasets) == 1

    index.datasets.archive([ls8_eo3_dataset2.id])

    datasets = list(index.datasets.search(archived=False, product=prod))
    assert len(datasets) == 0
    datasets = list(index.datasets.search(archived=None, product=prod))
    assert len(datasets) == 2
    datasets = list(index.datasets.search(archived=True, product=prod))
    assert len(datasets) == 2


def test_search_order_by_eo3(
    index: Index, ls8_eo3_dataset, ls8_eo3_dataset2, ls8_eo3_dataset3
) -> None:
    # provided as a string
    datasets = list(index.datasets.search(order_by=["id"]))
    assert len(datasets) == 3
    assert str(datasets[0].id) < str(datasets[1].id)
    assert str(datasets[1].id) < str(datasets[2].id)

    # provided as a Field
    prod = ls8_eo3_dataset.product
    fields = prod.metadata_type.dataset_fields
    index.datasets.archive([ls8_eo3_dataset3.id])
    datasets = list(index.datasets.search(order_by=[fields["id"]]))
    assert len(datasets) == 2
    assert str(datasets[0].id) < str(datasets[1].id)

    # ensure limit doesn't interfere with ordering
    datasets = list(index.datasets.search(order_by=["id"], limit=1))
    assert datasets[0] == ls8_eo3_dataset2

    datasets = list(
        index.datasets.search(order_by=[fields["id"].alchemy_expression.desc()])
    )
    assert len(datasets) == 2
    assert str(datasets[0].id) > str(datasets[1].id)


def test_search_or_expressions_eo3(
    index: Index,
    ls8_eo3_dataset: Dataset,
    ls8_eo3_dataset2: Dataset,
    wo_eo3_dataset: Dataset,
) -> None:
    # Three EO3 datasets:
    # - two landsat8 ard
    # - one wo

    all_datasets = list(index.datasets.search())
    assert len(all_datasets) == 3
    all_ids = {dataset.id for dataset in all_datasets}

    # OR all instruments: should return all datasets
    datasets = list(
        index.datasets.search(instrument=["WOOLI_TIRS", "OLI_TIRS", "OLI_TIRS2"])
    )
    assert len(datasets) == 3
    ids = {dataset.id for dataset in datasets}
    assert ids == all_ids

    # OR expression with only one clause.
    datasets = list(index.datasets.search(instrument=["OLI_TIRS"]))
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # OR both products: return all
    datasets = list(
        index.datasets.search(
            product=[ls8_eo3_dataset.product.name, wo_eo3_dataset.product.name]
        )
    )
    assert len(datasets) == 3
    ids = {dataset.id for dataset in datasets}
    assert ids == all_ids

    # eo OR eo3: return all
    datasets = list(
        index.datasets.search(
            metadata_type=[
                # LS5 + children
                ls8_eo3_dataset.metadata_type.name,
                # Nothing
                # LS8 dataset
                wo_eo3_dataset.metadata_type.name,
            ]
        )
    )
    assert len(datasets) == 3
    ids = {dataset.id for dataset in datasets}
    assert ids == all_ids

    # Redundant ORs should have no effect.
    datasets = list(
        index.datasets.search(
            product=[
                wo_eo3_dataset.product.name,
                wo_eo3_dataset.product.name,
                wo_eo3_dataset.product.name,
            ]
        )
    )
    assert len(datasets) == 1
    assert datasets[0].id == wo_eo3_dataset.id


def test_search_returning_eo3(
    index: Index,
    cfg_env: ODCEnvironment,
    ls8_eo3_dataset: Dataset,
    ls8_eo3_dataset2: Dataset,
    wo_eo3_dataset: Dataset,
) -> None:
    assert index.datasets.count() == 3, "Expected three test datasets"

    # Expect one product with our one dataset.
    results = list(
        index.datasets.search_returning(
            ("id", "region_code", "dataset_maturity"),
            platform="landsat-8",
            instrument="OLI_TIRS",
        )
    )
    assert len(results) == 1
    id_, region_code, maturity = results[0]
    assert id_ == ls8_eo3_dataset.id
    assert region_code == "090086"
    assert maturity == "final"

    count_by_date = index.datasets.count(
        product="ga_ls8c_ard_3",
        time=Range(
            begin=datetime.datetime(2016, 5, 12, 18, tzinfo=datetime.timezone.utc),
            end=datetime.datetime(2016, 5, 13, 2, tzinfo=datetime.timezone.utc),
        ),
    )
    assert count_by_date == 1
    results = list(
        index.datasets.search_returning(
            ("id", "metadata_doc"), platform="landsat-8", instrument="OLI_TIRS"
        )
    )
    assert len(results) == 1
    id_, document = results[0]
    assert id_ == ls8_eo3_dataset.id
    assert document == ls8_eo3_dataset.metadata_doc

    my_username = cfg_env.db_username
    if not my_username:
        my_username = _DEFAULT_DB_USER

    # Mixture of document and native fields
    results = list(
        index.datasets.search_returning(
            ("id", "creation_time", "format", "label"),
            platform="landsat-8",
            instrument="OLI_TIRS",
            indexed_by=my_username,
        )
    )
    assert len(results) == 1

    id_, creation_time, format_, label = results[0]

    assert id_ == ls8_eo3_dataset.id
    assert format_ == "GeoTIFF"

    # It's always UTC in the document
    expected_time = creation_time.astimezone(UTC).replace(tzinfo=None)
    assert expected_time.isoformat() == ls8_eo3_dataset.metadata.creation_dt
    assert label == ls8_eo3_dataset.metadata.label

    # All Fields
    results = list(index.datasets.search_returning(platform="landsat-8"))
    assert len(results) == 3

    assert ls8_eo3_dataset.id in (result.id for result in results)  # type: ignore[attr-defined]


def test_search_returning_rows_eo3(
    index: Index,
    eo3_ls8_dataset_doc,
    eo3_ls8_dataset2_doc,
    ls8_eo3_dataset,
    ls8_eo3_dataset2,
) -> None:
    dataset = ls8_eo3_dataset
    uri = eo3_ls8_dataset_doc[1]
    uri2 = eo3_ls8_dataset2_doc[1]
    results = list(
        index.datasets.search_returning(
            ("id", "uri"),
            platform="landsat-8",
            instrument="OLI_TIRS",
        )
    )
    assert len(results) == 1
    assert results == [(dataset.id, uri)]

    results = list(
        index.datasets.search_returning(
            ("id", "uri"),
            custom_offsets={
                "cloud_shadow": ("properties", "fmask:cloud_shadow"),
                "sun_azimuth": ("properties", "eo:sun_azimuth"),
            },
            platform="landsat-8",
            instrument="OLI_TIRS",
        )
    )
    assert len(results) == 1
    assert results[0].id == dataset.id  # type: ignore[attr-defined]
    assert 1.31 < results[0].cloud_shadow < 1.32  # type: ignore[attr-defined]
    assert 34.58 < results[0].sun_azimuth < 34.59  # type: ignore[attr-defined]

    results = list(
        index.datasets.search_returning(
            [],
            custom_offsets={
                "cloud_shadow": ("properties", "fmask:cloud_shadow"),
                "sun_azimuth": ("properties", "eo:sun_azimuth"),
            },
            platform="landsat-8",
            instrument="OLI_TIRS",
        )
    )
    assert len(results) == 1
    assert 1.31 < results[0].cloud_shadow < 1.32  # type: ignore[attr-defined]

    # A second dataset already has a location:
    results = set(
        index.datasets.search_returning(
            ("id", "uri"),
            platform="landsat-8",
            dataset_maturity="final",
        )
    )
    assert len(results) == 2
    assert results == {
        (dataset.id, uri),
        (ls8_eo3_dataset2.id, uri2),
    }


@pytest.mark.parametrize("datacube_env_name", ("postgis",))
def test_search_returning_uri(
    index: Index, eo3_ls8_dataset_doc, ls8_eo3_dataset
) -> None:
    dataset = ls8_eo3_dataset
    uri = eo3_ls8_dataset_doc[1]

    with suppress_deprecations():
        # If returning a field like uri, there will be one result per dataset.
        index.datasets.remove_location(dataset.id, uri)  # deprecated method
    results = list(
        index.datasets.search_returning(
            ("id", "uri"),
            platform="landsat-8",
            instrument="OLI_TIRS",
        )
    )
    assert len(results) == 1


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_search_returning_uris_legacy(
    index: Index,
    eo3_ls8_dataset_doc,
    eo3_ls8_dataset2_doc,
    ls8_eo3_dataset,
    ls8_eo3_dataset2,
) -> None:
    dataset = ls8_eo3_dataset
    uri = eo3_ls8_dataset_doc[1]
    uri3 = eo3_ls8_dataset2_doc[1]

    # If returning a field like uri, there will be one result per location.
    with suppress_deprecations():
        # No locations
        index.datasets.archive_location(dataset.id, uri)
        index.datasets.remove_location(dataset.id, uri)
        results = list(
            index.datasets.search_returning(
                ("id", "uri"),
                platform="landsat-8",
                instrument="OLI_TIRS",
            )
        )
        assert len(results) == 0

        # Add a second location and we should get two results
        index.datasets.add_location(dataset.id, uri)
        uri2 = "file:///tmp/test2"
        index.datasets.add_location(dataset.id, uri2)
        results = set(
            index.datasets.search_returning(
                ("id", "uri"),
                platform="landsat-8",
                instrument="OLI_TIRS",
            )
        )
    assert len(results) == 2
    assert results == {(dataset.id, uri), (dataset.id, uri2)}

    # A second dataset already has a location:
    results = set(
        index.datasets.search_returning(
            ("id", "uri"),
            platform="landsat-8",
            dataset_maturity="final",
        )
    )
    assert len(results) == 3
    assert results == {
        (dataset.id, uri),
        (dataset.id, uri2),
        (ls8_eo3_dataset2.id, uri3),
    }


def test_searches_only_type_eo3(
    index: Index, wo_eo3_dataset: Dataset, ls8_eo3_dataset: Dataset
) -> None:
    assert ls8_eo3_dataset.metadata_type.name != wo_eo3_dataset.metadata_type.name

    # One result in the product
    datasets = list(
        index.datasets.search(product=wo_eo3_dataset.product.name, platform="landsat-8")
    )
    assert len(datasets) == 1
    assert datasets[0].id == wo_eo3_dataset.id

    # One result in the metadata type
    datasets = list(index.datasets.search(metadata_type="eo3", platform="landsat-8"))
    assert len(datasets) == 1
    assert datasets[0].id == wo_eo3_dataset.id

    # No results when searching for a different dataset type.
    with pytest.raises(ValueError):
        next(index.datasets.search(product="spam_and_eggs", platform="landsat-8"))

    # Two result when no types specified.
    datasets = list(index.datasets.search(platform="landsat-8"))
    assert len(datasets) == 2
    assert {ds.id for ds in datasets} == {ls8_eo3_dataset.id, wo_eo3_dataset.id}

    # No results for different metadata type.
    with pytest.raises(ValueError):
        next(
            index.datasets.search(
                metadata_type="spam_type",
                platform="landsat-8",
            )
        )


def test_search_special_fields_eo3(
    index: Index, ls8_eo3_dataset: Dataset, wo_eo3_dataset: Dataset
) -> None:
    # 'product' is a special case
    datasets = list(index.datasets.search(product=ls8_eo3_dataset.product.name))
    assert len(datasets) == 1
    assert datasets[0].id == ls8_eo3_dataset.id

    # Unknown field: no results
    with pytest.raises(ValueError):
        next(
            index.datasets.search(
                platform="landsat-8",
                flavour="vanilla",
            )
        )


def test_search_by_uri_eo3(
    index: Index, ls8_eo3_dataset, ls8_eo3_dataset2, eo3_ls8_dataset_doc
) -> None:
    datasets = list(
        index.datasets.search(
            product=ls8_eo3_dataset.product.name, uri=eo3_ls8_dataset_doc[1]
        )
    )
    assert len(datasets) == 1
    datasets = list(
        index.datasets.search(product=ls8_eo3_dataset.product.name, uri="file:///x/yz")
    )
    assert len(datasets) == 0


def test_search_conflicting_types(index: Index, ls8_eo3_dataset) -> None:
    # Should return no results.
    with pytest.raises(ValueError):
        next(
            index.datasets.search(
                product=ls8_eo3_dataset.product.name,
                # The ls8 type is not of type storage_unit.
                metadata_type="storage_unit",
            )
        )


def test_fetch_all_of_md_type(index: Index, ls8_eo3_dataset: Dataset) -> None:
    # Get every dataset of the md type.
    assert ls8_eo3_dataset.metadata_type is not None  # to shut up mypy
    results = list(
        index.datasets.search(metadata_type=ls8_eo3_dataset.metadata_type.name)
    )
    assert len(results) == 1
    assert results[0].id == ls8_eo3_dataset.id
    # Get every dataset of the type.
    results = list(index.datasets.search(product=ls8_eo3_dataset.product.name))
    assert len(results) == 1
    assert results[0].id == ls8_eo3_dataset.id

    # No results for another.
    with pytest.raises(ValueError):
        next(index.datasets.search(metadata_type="spam_and_eggs"))


def test_count_searches(index: Index, ls8_eo3_dataset: Dataset) -> None:
    # One result in the telemetry type
    datasets = index.datasets.count(
        product=ls8_eo3_dataset.product.name,
        platform="landsat-8",
        instrument="OLI_TIRS",
    )
    assert datasets == 1

    # One result in the metadata type
    datasets = index.datasets.count(
        metadata_type=ls8_eo3_dataset.metadata_type.name,
        platform="landsat-8",
        instrument="OLI_TIRS",
    )
    assert datasets == 1

    # No results when searching for a different dataset type.
    datasets = index.datasets.count(
        product="spam_and_eggs", platform="landsat-8", instrument="OLI_TIRS"
    )
    assert datasets == 0

    # One result when no types specified.
    datasets = index.datasets.count(
        platform="landsat-8",
        instrument="OLI_TIRS",
    )
    assert datasets == 1

    # No results for different metadata type.
    datasets = index.datasets.count(
        metadata_type="spam_and_eggs", platform="landsat-8", instrument="OLI_TIRS"
    )
    assert datasets == 0


def test_count_by_product_searches_eo3(
    index: Index,
    ls8_eo3_dataset: Dataset,
    ls8_eo3_dataset2: Dataset,
    wo_eo3_dataset: Dataset,
) -> None:
    # Two result in the ls8 type
    products = tuple(
        index.datasets.count_by_product(
            product=ls8_eo3_dataset.product.name, platform="landsat-8"
        )
    )
    assert products == ((ls8_eo3_dataset.product, 2),)

    # Two results in the metadata type
    products = tuple(
        index.datasets.count_by_product(
            metadata_type=ls8_eo3_dataset.metadata_type.name,
            platform="landsat-8",
        )
    )
    assert products == ((ls8_eo3_dataset.product, 2),)

    # No results when searching for a different dataset type.
    products = tuple(
        index.datasets.count_by_product(product="spam_and_eggs", platform="landsat-8")
    )
    assert products == ()

    # Three results over 2 products when no types specified.
    products = set(
        index.datasets.count_by_product(
            platform="landsat-8",
        )
    )
    assert products == {(ls8_eo3_dataset.product, 2), (wo_eo3_dataset.product, 1)}

    # No results for different metadata type.
    products = tuple(
        index.datasets.count_by_product(
            metadata_type="spam_and_eggs",
        )
    )
    assert products == ()

    index.datasets.archive([ls8_eo3_dataset.id])
    products = tuple(
        index.datasets.count_by_product(
            product=ls8_eo3_dataset.product.name, platform="landsat-8"
        )
    )
    assert products == ((ls8_eo3_dataset.product, 1),)
    products = tuple(
        index.datasets.count_by_product(
            archived=True, product=ls8_eo3_dataset.product.name, platform="landsat-8"
        )
    )
    assert products == ((ls8_eo3_dataset.product, 1),)
    products = tuple(
        index.datasets.count_by_product(
            archived=None, product=ls8_eo3_dataset.product.name, platform="landsat-8"
        )
    )
    assert products == ((ls8_eo3_dataset.product, 2),)


def test_count_time_groups(index: Index, ls8_eo3_dataset: Dataset) -> None:
    timeline = list(
        index.datasets.count_product_through_time(
            "1 day",
            product=ls8_eo3_dataset.product.name,
            time=Range(
                datetime.datetime(2016, 5, 11, tzinfo=UTC),
                datetime.datetime(2016, 5, 13, tzinfo=UTC),
            ),
        )
    )

    assert len(timeline) == 2
    assert timeline == [
        (
            Range(
                datetime.datetime(2016, 5, 11, tzinfo=UTC),
                datetime.datetime(2016, 5, 12, tzinfo=UTC),
            ),
            0,
        ),
        (
            Range(
                datetime.datetime(2016, 5, 12, tzinfo=UTC),
                datetime.datetime(2016, 5, 13, tzinfo=UTC),
            ),
            1,
        ),
    ]


def test_count_time_groups_cli(clirunner: Any, ls8_eo3_dataset: Dataset) -> None:
    result = clirunner(
        ["product-counts", "1 day", "time in [2016-05-11, 2016-05-13]"],
        cli_method=datacube.scripts.search_tool.cli,
        verbose_flag="",
    )
    expected_out = (
        f"{ls8_eo3_dataset.product.name}\n    2016-05-11: 0\n    2016-05-12: 1\n"
    )
    assert result.stdout.endswith(expected_out)

    # updated version of the test
    result = clirunner(
        [
            "dataset",
            "count",
            "--period",
            "1 day",
            "--query",
            "time in [2016-05-11, 2016-05-13]",
        ],
        cli_method=datacube.scripts.cli_app.cli,
        verbose_flag="",
    )
    assert result.output.endswith(
        f"product: {ls8_eo3_dataset.product.name}\ntime: '2016-05-12'\ncount: 1\n"
    )


def test_search_cli_basic(clirunner: Any, ls8_eo3_dataset: Dataset) -> None:
    """
    Search datasets using the cli.
    """
    result = clirunner(
        [
            # No search arguments: return all datasets.
            "datasets"
        ],
        cli_method=datacube.scripts.search_tool.cli,
    )
    assert str(ls8_eo3_dataset.id) in result.output
    assert str(ls8_eo3_dataset.metadata_type.name) in result.output
    assert result.exit_code == 0

    result = clirunner(
        ["dataset", "search"],
        cli_method=datacube.scripts.cli_app.cli,
    )
    assert str(ls8_eo3_dataset.id) in result.output
    assert str(ls8_eo3_dataset.product.name) in result.output
    assert result.exit_code == 0


def test_cli_info_eo3(
    index: Index,
    clirunner: Any,
    ls8_eo3_dataset: Dataset,
    ls8_eo3_dataset2: Dataset,
    eo3_ls8_dataset_doc,
) -> None:
    """
    Search datasets using the cli.
    """
    opts = ["dataset", "info", str(ls8_eo3_dataset.id)]
    with suppress_deprecations():
        result = clirunner(opts, verbose_flag="")

    output = result.output
    output_lines = list(output.splitlines())

    # Should be a valid yaml
    yaml_docs = list(yaml.safe_load_all(output))
    assert len(yaml_docs) == 1

    # We output properties in order for readability:
    output_lines = set(output_lines)
    expected_lines = [
        "id: " + str(ls8_eo3_dataset.id),
        "product: ga_ls8c_ard_3",
        "status: active",
        "location: " + str(ls8_eo3_dataset.uri),
        "fields:",
        "    creation_time: 2019-10-07 20:19:19.218290",
        "    format: GeoTIFF",
        "    instrument: OLI_TIRS",
        "    label: ga_ls8c_ard_3-0-0_090086_2016-05-12_final",
        "    landsat_product_id: LC08_L1TP_090086_20160512_20180203_01_T1",
        "    landsat_scene_id: LC80900862016133LGN02",
        "    lat: {begin: -38.53221689818913, end: -36.41618895501644}",
        "    lon: {begin: 147.65992717003462, end: 150.3003802932316}",
        "    platform: landsat-8",
        "    product_family: ard",
        "    region_code: 090086",
        "    time: {begin: '2016-05-12T23:50:23.054165+00:00', end: '2016-05-12T23:50:52.031499+00:00'}",
    ]
    for line in expected_lines:
        assert line in output_lines

    # Check indexed time separately, as we don't care what timezone it's displayed in.
    indexed_time = yaml_docs[0]["indexed"]
    assert isinstance(indexed_time, datetime.datetime)
    t = ls8_eo3_dataset.indexed_time
    assert t is not None
    assert tz_as_utc(indexed_time) == tz_as_utc(t)

    # Request two, they should have separate yaml documents
    opts.append(str(ls8_eo3_dataset2.id))

    result = clirunner(opts)
    yaml_docs = list(yaml.safe_load_all(result.stdout))
    assert len(yaml_docs) == 2, "Two datasets should produce two sets of info"
    assert yaml_docs[0]["id"] == str(ls8_eo3_dataset.id)
    assert yaml_docs[1]["id"] == str(ls8_eo3_dataset2.id)


def test_find_duplicates_eo3(
    index,
    ls8_eo3_dataset,
    ls8_eo3_dataset2,
    ls8_eo3_dataset3,
    ls8_eo3_dataset4,
    wo_eo3_dataset,
) -> None:
    # Our four ls8 datasets and one wo.
    all_datasets = list(index.datasets.search())
    assert len(all_datasets) == 5

    # First two ls8 datasets have the same path/row, last two have a different row.
    search_result = namedtuple("search_result", ["region_code", "dataset_maturity"])
    expected_ls8_path_row_duplicates = [
        (search_result("090086", "final"), {ls8_eo3_dataset.id, ls8_eo3_dataset2.id}),
        (search_result("101077", "final"), {ls8_eo3_dataset3.id, ls8_eo3_dataset4.id}),
    ]

    # Specifying groups as fields:
    f = ls8_eo3_dataset.metadata_type.dataset_fields.get
    field_res = sorted(
        index.datasets.search_product_duplicates(
            ls8_eo3_dataset.product, f("region_code"), f("dataset_maturity")
        )
    )
    assert field_res == expected_ls8_path_row_duplicates
    # Field names as strings
    product_res = sorted(
        index.datasets.search_product_duplicates(
            ls8_eo3_dataset.product, "region_code", "dataset_maturity"
        )
    )
    assert product_res == expected_ls8_path_row_duplicates

    # No WO duplicates: there's only one
    sat_res = sorted(
        index.datasets.search_product_duplicates(
            wo_eo3_dataset.product, "region_code", "dataset_maturity"
        )
    )
    assert sat_res == []


def test_find_duplicates_with_time(
    index: Index, nrt_dataset, final_dataset, ls8_eo3_dataset
) -> None:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", FixWindingWarning)
        index.datasets.add(nrt_dataset, with_lineage=False)
        index.datasets.add(final_dataset, with_lineage=False)
    d = index.datasets.get(nrt_dataset.id)
    assert d is not None
    assert not d.is_archived
    d = index.datasets.get(final_dataset.id)
    assert d is not None
    assert not d.is_archived

    all_datasets = list(index.datasets.search())
    assert len(all_datasets) == 3

    search_result = namedtuple("search_result", ["region_code", "time"])

    expected_result_old = [
        (
            search_result(
                "090086", '("2023-04-30 23:50:33.884549","2023-04-30 23:50:34.884549")'
            ),
            {nrt_dataset.id, final_dataset.id},
        )
    ]
    expected_result_new = [
        (
            search_result(
                "090086",
                (
                    datetime.datetime(
                        2023, 4, 30, 23, 50, 33, 884549, tzinfo=datetime.timezone.utc
                    ),
                    datetime.datetime(
                        2023, 4, 30, 23, 50, 34, 884549, tzinfo=datetime.timezone.utc
                    ),
                ),
            ),
            {nrt_dataset.id, final_dataset.id},
        )
    ]
    res = sorted(
        index.datasets.search_product_duplicates(
            nrt_dataset.product,
            "region_code",
            "time",
        )
    )

    assert res in (expected_result_old, expected_result_new)


def test_csv_search_via_cli_eo3(
    clirunner: Any, ls8_eo3_dataset: Dataset, ls8_eo3_dataset2: Dataset
) -> None:
    """
    Search datasets via the cli with csv output
    """

    def matches_both(*args) -> None:
        rows = _cli_csv_search(("datasets",) + args, clirunner)
        assert len(rows) == 2
        assert {rows[0]["id"], rows[1]["id"]} == {
            str(ls8_eo3_dataset.id),
            str(ls8_eo3_dataset2.id),
        }

    def matches_1(*args) -> None:
        rows = _cli_csv_search(("datasets",) + args, clirunner)
        assert len(rows) == 1
        assert rows[0]["id"] == str(ls8_eo3_dataset.id)

    def matches_none(*args) -> None:
        rows = _cli_csv_search(("datasets",) + args, clirunner)
        assert len(rows) == 0

    def no_such_product(*args) -> None:
        with pytest.raises(ValueError):
            _cli_csv_search(("datasets",) + args, clirunner)

    matches_both("lat in [-40, -10]")
    matches_both("product=" + ls8_eo3_dataset.product.name)

    # Don't return on a mismatch
    matches_none("lat in [150, 160]")

    # Match only a single dataset using multiple fields
    matches_1("platform=landsat-8", "time in [2016-05-11, 2016-05-13]")

    # One matching field, one non-matching
    no_such_product("time in [2016-05-11, 2014-05-13]", "platform=landsat-5")

    # Test date shorthand
    matches_both("time in [2016-05, 2016-05]")
    matches_none("time in [2014-06, 2014-06]")

    matches_both("time in 2016-05")
    matches_none("time in 2014-08")
    matches_both("time in 2016")
    matches_none("time in 2015")

    matches_both("time in [2016, 2016]")
    matches_both("time in [2015, 2017]")
    matches_none("time in [2015, 2015]")
    matches_none("time in [2013, 2013]")

    matches_both("time in [2016-4, 2016-8]")
    matches_none("time in [2016-1, 2016-3]")
    matches_both("time in [2005, 2017]")


_EXT_AND_BASE_EO3_OUTPUT_HEADER = [
    "id",
    "crs_raw",
    "dataset_maturity",
    "eo_gsd",
    "eo_sun_azimuth",
    "eo_sun_elevation",
    "cloud_cover",
    "fmask_clear",
    "fmask_cloud_shadow",
    "fmask_snow",
    "fmask_water",
    "format",
    "gqa",
    "gqa_abs_iterative_mean_x",
    "gqa_abs_iterative_mean_xy",
    "gqa_abs_iterative_mean_y",
    "gqa_abs_x,gqa_abs_xy",
    "gqa_abs_y",
    "gqa_cep90",
    "gqa_iterative_mean_x",
    "gqa_iterative_mean_xy",
    "gqa_iterative_mean_y",
    "gqa_iterative_stddev_x",
    "gqa_iterative_stddev_xy",
    "gqa_iterative_stddev_y",
    "gqa_mean_x",
    "gqa_mean_xy",
    "gqa_mean_y,gqa_stddev_x",
    "gqa_stddev_xy",
    "gqa_stddev_y",
    "creation_time",
    "indexed_by",
    "indexed_time",
    "instrument",
    "label",
    "landsat_product_id",
    "landsat_scene_id",
    "lat",
    "lon",
    "metadata_doc",
    "metadata_type",
    "metadata_type_id",
    "platform",
    "product",
    "product_family",
    "region_code",
    "time",
    "uri",
]


def test_csv_structure_eo3(clirunner, ls8_eo3_dataset, ls8_eo3_dataset2) -> None:
    output = _csv_search_raw(["datasets", " lat in [-40, -10]"], clirunner)
    lines = [line.strip() for line in output.split("\n") if line]
    # A header and two dataset rows
    assert len(lines) == 3
    header_line = lines[0]
    for header in _EXT_AND_BASE_EO3_OUTPUT_HEADER:
        assert header in header_line


def test_query_dataset_multi_product_eo3(
    index: Index, ls8_eo3_dataset, wo_eo3_dataset
) -> None:
    # We have one ls5 level1 and its child nbar
    dc = Datacube(index)

    # Can we query a single product name?
    datasets = dc.find_datasets(product="ga_ls8c_ard_3")
    assert len(datasets) == 1

    # Can we query multiple products?
    datasets = dc.find_datasets(product=["ga_ls8c_ard_3", "ga_ls_wo_3"])
    assert len(datasets) == 2

    # Can we query multiple products in a tuple
    datasets = dc.find_datasets(product=("ga_ls8c_ard_3", "ga_ls_wo_3"))
    assert len(datasets) == 2


def test_search_boolean_eo3(index: Index, s1_eo3_dataset) -> None:
    res = list(
        index.datasets.search(
            product=s1_eo3_dataset.product.name, speckle_filter_applied=False
        )
    )
    assert len(res) == 1
    res = list(
        index.datasets.search(
            product=s1_eo3_dataset.product.name, speckle_filter_applied=True
        )
    )
    assert len(res) == 0



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest

from datacube.utils.changes import (
    MISSING,
    Change,
    allow_addition,
    allow_any,
    allow_extension,
    allow_removal,
    allow_truncation,
    classify_changes,
    contains,
)


def test_changes_contains() -> None:
    assert contains("bob", "BOB") is True
    assert contains("bob", "BOB", case_sensitive=True) is False
    assert contains(1, 1) is True
    assert contains(1, {}) is False
    # same as above, but with None interpreted as {}
    assert contains(1, None) is False
    assert contains({}, 1) is False
    assert contains(None, 1) is False
    assert contains({}, {}) is True
    assert contains({}, None) is True

    # this one is arguable...
    assert contains(None, {}) is False
    assert contains(None, None) is True
    assert contains({"a": 1, "b": 2}, {"a": 1}) is True
    assert contains({"a": {"b": "BOB"}}, {"a": {"b": "bob"}}) is True
    assert (
        contains({"a": {"b": "BOB"}}, {"a": {"b": "bob"}}, case_sensitive=True) is False
    )
    assert contains("bob", "alice") is False
    assert contains({"a": 1}, {"a": 1, "b": 2}) is False
    assert contains({"a": {"b": 1}}, {"a": {}}) is True
    assert contains({"a": {"b": 1}}, {"a": None}) is True


def test_classify_changes() -> None:
    assert classify_changes([], {}) == ([], [])
    assert classify_changes([(("a",), 1, 2)], {}) == ([], [(("a",), 1, 2)])
    assert classify_changes([(("a",), 1, 2)], {("a",): allow_any}) == (
        [(("a",), 1, 2)],
        [],
    )

    changes: list[Change] = [
        (("a2",), {"b1": 1}, MISSING)
    ]  # {'a1': 1, 'a2': {'b1': 1}} → {'a1': 1}
    good_change: tuple[list[Change], list[Change]] = (changes, [])
    bad_change: tuple[list[Change], list[Change]] = ([], changes)
    assert classify_changes(changes, {}) == bad_change
    assert classify_changes(changes, {(): allow_any}) == good_change
    assert classify_changes(changes, {(): allow_removal}) == bad_change
    assert classify_changes(changes, {(): allow_addition}) == bad_change
    assert classify_changes(changes, {(): allow_truncation}) == good_change
    assert classify_changes(changes, {(): allow_extension}) == bad_change
    assert classify_changes(changes, {("a1",): allow_any}) == bad_change
    assert classify_changes(changes, {("a1",): allow_removal}) == bad_change
    assert classify_changes(changes, {("a1",): allow_addition}) == bad_change
    assert classify_changes(changes, {("a1",): allow_truncation}) == bad_change
    assert classify_changes(changes, {("a1",): allow_extension}) == bad_change
    assert classify_changes(changes, {("a2",): allow_any}) == good_change
    assert classify_changes(changes, {("a2",): allow_removal}) == good_change
    assert classify_changes(changes, {("a2",): allow_addition}) == bad_change
    assert classify_changes(changes, {("a2",): allow_truncation}) == bad_change
    assert classify_changes(changes, {("a2",): allow_extension}) == bad_change

    # Test that type errors result in RuntimeError.
    with pytest.raises(RuntimeError):
        classify_changes(changes, {("a2",): object()})  # type: ignore[dict-item]

    assert str(MISSING) == repr(MISSING)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest
from sqlalchemy.engine.url import URL
from sqlalchemy.exc import OperationalError

from datacube.drivers.postgres._connections import (
    PostgresDb,
    handle_dynamic_token_authentication,
)

counter = [0]
last_base = [None]


def next_token(base) -> str:
    counter[0] += 1
    last_base[0] = base
    return f"{base}{counter[0]}"


def test_dynamic_password() -> None:
    url = URL.create(
        "postgresql",
        host="fake_host",
        database="fake_database",
        port=6543,
        username="fake_username",
        password="fake_password",
    )
    engine = PostgresDb._create_engine(url)
    counter[0] = 0
    last_base[0] = None
    handle_dynamic_token_authentication(engine, next_token, base="password")
    with pytest.raises(OperationalError):
        engine.connect()
    assert counter[0] == 1
    assert last_base[0] == "password"



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging

import click
from click import echo, style
from sqlalchemy.exc import OperationalError

import datacube
from datacube.cfg import ODCEnvironment, psql_url_from_config
from datacube.drivers.postgres._connections import IndexSetupError
from datacube.index import index_connect
from datacube.ui import click as ui
from datacube.ui.click import cli, handle_exception

_LOG: logging.Logger = logging.getLogger("datacube-system")


@cli.group(name="system", help="System commands")
def system() -> None:
    pass


@system.command("init", help="Initialise the database")
@click.option(
    "--default-types/--no-default-types",
    is_flag=True,
    default=True,
    help="Add default types? (default: true)",
)
@click.option(
    "--init-users/--no-init-users",
    is_flag=True,
    default=True,
    help="Include user roles and grants. (default: true)",
)
@click.option(
    "--recreate-views/--no-recreate-views",
    is_flag=True,
    default=True,
    help="Recreate dynamic views",
)
@click.option(
    "--rebuild/--no-rebuild",
    is_flag=True,
    default=False,
    help="Rebuild all dynamic fields (caution: slow)",
)
@click.option(
    "--lock-table/--no-lock-table",
    is_flag=True,
    default=False,
    help="Allow table to be locked (eg. while creating missing indexes)",
)
@ui.pass_index(expect_initialised=False)
def database_init(
    index, default_types, init_users, recreate_views, rebuild, lock_table
) -> None:
    echo("Initialising database...")

    was_created = index.init_db(
        with_default_types=default_types, with_permissions=init_users
    )

    if was_created:
        echo(style("Created.", bold=True))
    else:
        echo(style("Updated.", bold=True))

    echo("Checking indexes/views.")
    index.metadata_types.check_field_indexes(
        allow_table_lock=lock_table,
        rebuild_indexes=rebuild,
        rebuild_views=recreate_views or rebuild,
    )
    echo("Done.")


@system.command("check", help="Check and display current configuration")
@ui.pass_config
def check(cfg_env: ODCEnvironment) -> None:
    """
    Verify & view current configuration
    """

    def echo_field(name: str, value) -> None:
        echo("{:<15}".format(name + ":") + style(str(value), bold=True))

    echo_field("Version", datacube.__version__)
    echo_field("Index Driver", cfg_env.index_driver)
    db_url = psql_url_from_config(cfg_env)
    echo_field("Database URL:", db_url)
    echo_field("Skip broken datasets:", cfg_env.skip_broken_datasets)
    echo_field("Datacube load limit:", cfg_env.dc_load_limit)

    echo()
    echo("Valid connection:\t", nl=False)
    try:
        index = index_connect(config_env=cfg_env)
        echo(style("YES", bold=True))
        if index.url_parts.username:
            for role, user, _ in index.users.list_users():
                if user == index.url_parts.username:
                    echo(f"You have {style(role.upper(), bold=True)} privileges.")
    except OperationalError as e:
        handle_exception("Error Connecting to Database: %s", e)
    except IndexSetupError as e:
        handle_exception("Database not initialised: %s", e)


@system.command("clone", help="Clone an existing ODC index into a new, empty index")
@click.option(
    "--batch-size",
    help="Size of batches for bulk-adding to the new index",
    type=int,
    default=1000,
)
@click.option(
    "--skip-lineage/--no-skip-lineage",
    is_flag=True,
    default=False,
    help="Do not load lineage data where possible. (default: false - i.e. do not skip lineage)",
)
@click.option(
    "--lineage-only/--no-lineage-only",
    is_flag=True,
    default=False,
    help="Clone lineage data only. (default: false)",
)
@click.argument("source-env", type=str, nargs=1)
@ui.pass_config
def clone(
    env: ODCEnvironment,
    batch_size: int,
    skip_lineage: bool,
    lineage_only: bool,
    source_env: str,
) -> None:
    if skip_lineage and lineage_only:
        echo("Cannot set both lineage-only and skip-lineage", err=True)
        exit(1)
    try:
        destination_index = index_connect(env, validate_connection=True)
    except OperationalError as e:
        handle_exception("Error Connecting to Destination Database: %s", e)
        exit(1)
    except IndexSetupError as e:
        handle_exception("Destination database not initialised: %s", e)
        exit(1)

    try:
        source_config = env._cfg[source_env]
    except KeyError:
        raise click.ClickException(
            f"No datacube config found for '{source_env}'"
        ) from None

    try:
        src_index = index_connect(source_config, validate_connection=True)
    except OperationalError as e:
        handle_exception("Error Connecting to Source Database: %s", e)
        exit(1)
    except IndexSetupError as e:
        handle_exception("Source database not initialised: %s", e)
        exit(1)
    # Any errors will have be logged.
    destination_index.clone(
        src_index,
        batch_size=batch_size,
        skip_lineage=skip_lineage,
        lineage_only=lineage_only,
    )
    exit(0)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Methods for managing dynamic dataset field indexes and views.
"""

import logging
from collections.abc import Sequence

from sqlalchemy import Index, select, text

from ._core import schema_qualified
from ._schema import DATASET, METADATA_TYPE, PRODUCT
from .sql import CreateView, pg_exists

_LOG: logging.Logger = logging.getLogger(__name__)


def contains_all(d_: dict, *keys) -> bool:
    """
    Does the dictionary have values for all of the given keys?

    >>> contains_all({'a': 4}, 'a')
    True
    >>> contains_all({'a': 4, 'b': 5}, 'a', 'b')
    True
    >>> contains_all({'b': 5}, 'a')
    False
    """
    return all(d_.get(key) for key in keys)


def _ensure_view(
    conn, fields, name: str, replace_existing, where_expression, delete: bool = False
) -> None:
    """
    Ensure a view exists for the given fields
    """
    # Create a view of search fields (for debugging convenience).
    # 'dv_' prefix: dynamic view. To distinguish from views that are created as part of the schema itself.
    view_name = schema_qualified(f"dv_{name}_dataset")
    exists = pg_exists(conn, view_name)
    # This currently leaves a window of time without the views: it's primarily intended for development.
    if exists and replace_existing:
        _LOG.debug("Dropping view: %s (replace=%r)", view_name, replace_existing)
        conn.execute(text(f"drop view {view_name}"))
        exists = False
    if not exists:
        _LOG.debug("Creating view: %s", view_name)
        select_fields = [
            field.alchemy_expression.label(field.name)
            for field in fields.values()
            if not field.affects_row_selection
        ]
        conn.execute(
            CreateView(
                view_name,
                select(*select_fields)
                .select_from(DATASET.join(PRODUCT).join(METADATA_TYPE))
                .where(where_expression),
            )
        )
    elif exists and delete:
        _LOG.debug(f"Dropping view: {view_name}")
        conn.execute(text(f"drop view {view_name}"))
    else:
        _LOG.debug("View exists: %s (replace=%r)", view_name, replace_existing)
    legacy_name = schema_qualified(f"{name}_dataset")
    if pg_exists(conn, legacy_name):
        _LOG.debug("Dropping legacy view: %s", legacy_name)
        conn.execute(text(f"drop view {legacy_name}"))


def check_dynamic_fields(
    conn,
    concurrently,
    dataset_filter,
    excluded_field_names: Sequence[str],
    fields,
    name: str,
    rebuild_indexes: bool = False,
    rebuild_view: bool = False,
    delete_view: bool = False,
) -> None:
    """
    Check that we have expected indexes and views for the given fields
    """
    # If this type has time/space fields, create composite indexes (as they are often searched together)
    # We will probably move these into product configuration in the future.
    composite_indexes = (
        ("lat", "lon", "time"),
        ("time", "lat", "lon"),
        ("sat_path", "sat_row", "time"),
    )

    all_exclusions = tuple(excluded_field_names)
    for composite_names in composite_indexes:
        # If all of the fields are available in this product, we'll create a composite index
        # for them instead of individual indexes.
        if contains_all(fields, *composite_names):
            all_are_excluded = set(excluded_field_names) >= set(composite_names)
            _check_field_index(
                conn,
                [fields.get(f) for f in composite_names],
                name,
                dataset_filter,
                concurrently=concurrently,
                replace_existing=rebuild_indexes,
                # If all fields were excluded individually it should be removed.
                should_exist=not all_are_excluded,
                index_type="gist",
            )
            all_exclusions += composite_names

    # Create indexes for the individual fields.
    for field in fields.values():
        if not field.postgres_index_type:
            continue
        _check_field_index(
            conn,
            [field],
            name,
            dataset_filter,
            should_exist=field.indexed and (field.name not in all_exclusions),
            concurrently=concurrently,
            replace_existing=rebuild_indexes,
        )
    # A view of all fields
    _ensure_view(conn, fields, name, rebuild_view, dataset_filter, delete_view)


def _check_field_index(
    conn,
    fields,
    name_prefix,
    filter_expression,
    should_exist: bool = True,
    concurrently: bool = False,
    replace_existing: bool = False,
    index_type=None,
) -> None:
    """
    Check the status of a given index: add or remove it as needed
    """
    if index_type is None:
        if len(fields) > 1:
            raise ValueError("Must specify index type for composite indexes.")
        index_type = fields[0].postgres_index_type

    field_name = "_".join([f.name.lower() for f in fields])
    # Our normal indexes start with "ix_", dynamic indexes with "dix_"
    index_name = f"dix_{name_prefix.lower()}_{field_name}"
    # Previous naming scheme
    legacy_name = f"dix_field_{name_prefix.lower()}_dataset_{field_name}"
    indexed_expressions = [f.alchemy_expression for f in fields]
    index = Index(
        index_name,
        *indexed_expressions,
        postgresql_where=filter_expression,
        postgresql_using=index_type,
        # Don't lock the table (in the future we'll allow indexing new fields...)
        postgresql_concurrently=concurrently,
    )
    exists = pg_exists(conn, schema_qualified(index_name))
    legacy_exists = pg_exists(conn, schema_qualified(legacy_name))

    # This currently leaves a window of time without indexes: it's primarily intended for development.
    if replace_existing or (not should_exist):
        if exists:
            _LOG.debug("Dropping index: %s (replace=%r)", index_name, replace_existing)
            index.drop(conn)
            exists = False
        if legacy_exists:
            _LOG.debug(
                "Dropping legacy index: %s (replace=%r)", legacy_name, replace_existing
            )
            Index(legacy_name, *indexed_expressions).drop(conn)
            legacy_exists = False

    if should_exist:
        if not (exists or legacy_exists):
            _LOG.info("Creating index: %s", index_name)
            index.create(conn)
        else:
            _LOG.debug("Index exists: %s  (replace=%r)", index_name, replace_existing)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""rasterio environment management tools"""

import threading
from types import SimpleNamespace
from typing import Literal

import rasterio
import rasterio.env
from deprecat import deprecat
from rasterio.session import AWSSession, DummySession

from datacube.migration import ODC2DeprecationWarning
from datacube.utils.generic import thread_local_cache

_CFG_LOCK = threading.Lock()
_CFG = SimpleNamespace(aws=None, cloud_defaults=False, kwargs={}, epoch=0)


SECRET_KEYS = ("AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY", "AWS_SESSION_TOKEN")


def _sanitize(opts, keys):
    return {k: (v if k not in keys else "xx..xx") for k, v in opts.items()}


def _state(purge: bool = False):
    """
    .env   None| rasterio.Env
    .epoch -1  | +Int
    """
    return thread_local_cache(
        "__rio_state__", SimpleNamespace(env=None, epoch=-1), purge=purge
    )


def get_rio_env(sanitize: bool = True) -> dict:
    """Get GDAL params configured by rasterio for the current thread.

    :param sanitize: If True replace sensitive Values with 'x'
    """
    env = rasterio.env.local._env  # pylint: disable=protected-access
    if env is None:
        return {}
    opts = env.get_config_options()
    if sanitize:
        opts = _sanitize(opts, SECRET_KEYS)

    return opts


def deactivate_rio_env() -> None:
    """Exit previously configured environment, or do nothing if one wasn't configured."""
    state = _state(purge=True)

    if state.env is not None:
        state.env.__exit__(None, None, None)


def activate_rio_env(
    aws: Literal["auto"] | dict | None = None, cloud_defaults: bool = False, **kwargs
) -> dict:
    """Inject activated rasterio.Env into current thread.

    This de-activates previously setup environment.

    :param aws: Dictionary of options for rasterio.session.AWSSession
                OR 'auto' -- session = rasterio.session.AWSSession()

    :param cloud_defaults: When True inject settings for reading COGs
    :param kwargs: Passed on to rasterio.Env(..) constructor
    """
    session = DummySession()

    if aws is not None:
        if not (aws == "auto" or isinstance(aws, dict)):
            raise ValueError('Only support: None|"auto"|{..} for `aws` parameter')

        aws = {} if aws == "auto" else dict(**aws)
        region_name = aws.get("region_name", "auto")

        if region_name == "auto":
            from datacube.utils.aws import auto_find_region

            try:
                aws["region_name"] = auto_find_region()
            except ValueError as e:
                # only treat it as error if it was requested by user
                if "region_name" in aws:
                    raise e

        session = AWSSession(**aws)

    opts = (
        {
            "GDAL_DISABLE_READDIR_ON_OPEN": "EMPTY_DIR",
            "GDAL_HTTP_MAX_RETRY": "10",
            "GDAL_HTTP_RETRY_DELAY": "0.5",
        }
        if cloud_defaults
        else {}
    )

    opts.update(**kwargs)

    state = _state()

    if state.env is not None:
        state.env.__exit__(None, None, None)

    env = rasterio.Env(session=session, **opts)
    env.__enter__()
    state.env = env
    state.epoch = -1

    return get_rio_env()


def activate_from_config():
    """Check if this threads needs to reconfigure, then does reconfigure.

    - Does nothing if this thread is already configured and configuration hasn't changed.
    - Configures current thread with default rio settings
    """
    cfg = _CFG
    state = _state()

    if cfg.epoch != state.epoch:
        ee = activate_rio_env(
            aws=cfg.aws, cloud_defaults=cfg.cloud_defaults, **cfg.kwargs
        )
        state.epoch = cfg.epoch
        return ee

    return None


def set_default_rio_config(aws=None, cloud_defaults: bool = False, **kwargs) -> None:
    """Setup default configuration for rasterio/GDAL.

    Doesn't actually activate one, just stores configuration for future
    use from IO threads.

    :param aws: Dictionary of options for rasterio.session.AWSSession
                OR 'auto' -- session = rasterio.session.AWSSession()

    :param cloud_defaults: When True inject settings for reading COGs
    :param kwargs: Passed on to rasterio.Env(..) constructor
    """
    global _CFG  # pylint: disable=global-statement

    with _CFG_LOCK:
        _CFG = SimpleNamespace(
            aws=aws, cloud_defaults=cloud_defaults, kwargs=kwargs, epoch=_CFG.epoch + 1
        )


@deprecat(
    reason="This method is deprecated. Use datacube.utils.aws.configure_s3_access instead.",
    version="1.9.7",
    category=ODC2DeprecationWarning,
)
def configure_s3_access(
    profile: str | None = None,
    region_name: str = "auto",
    aws_unsigned: bool = False,
    requester_pays: bool = False,
    cloud_defaults: bool = True,
    client=None,
    **gdal_opts,
):
    """
    Use :meth:`datacube.utils.aws.configure_s3_access` instead.
    """
    from datacube.utils.aws import configure_s3_access

    return configure_s3_access(
        profile=profile,
        region_name=region_name,
        aws_unsigned=aws_unsigned,
        requester_pays=requester_pays,
        cloud_defaults=cloud_defaults,
        client=client,
        **gdal_opts,
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
odc.loader based load.

separate file to reduce formatting issues.

"""

from __future__ import annotations

from collections.abc import Callable, Generator, Mapping, Sequence
from datetime import datetime
from types import SimpleNamespace
from typing import Literal

import xarray as xr
from odc.geo.geobox import GeoBox, GeoboxTiles
from odc.loader import (
    FixedCoord,
    RasterBandMetadata,
    RasterGroupMetadata,
    RasterLoadParams,
    RasterSource,
    chunked_load,
    reader_driver,
    resolve_chunk_shape,
)
from odc.loader.types import ReaderDriverSpec

from datacube.storage._load import ProgressFunction

from ..model import Dataset, ExtraDimensions, Measurement
from . import BandInfo


def ds_geobox(ds: Dataset, **kw) -> GeoBox | None:
    from ..testutils.io import eo3_geobox

    try:
        return eo3_geobox(ds, **kw)
    except ValueError:
        return None


def _extract_coords(extra_dims: ExtraDimensions) -> list[FixedCoord]:
    coords = extra_dims.dims

    return [
        FixedCoord(
            name=k,
            values=d["values"],
            dtype=str(d.get("dtype", "float32")),
            dim=k,
            units=d.get("units", "1"),
        )
        for k, d in coords.items()
    ]


def driver_based_load(
    driver: ReaderDriverSpec,
    sources: xr.DataArray,
    geobox: GeoBox,
    measurements: Sequence[Measurement],
    dask_chunks: Mapping[str, int | Literal["auto"]] | None = None,
    skip_broken_datasets: bool = False,
    progress_cbk: ProgressFunction | None = None,
    extra_dims: ExtraDimensions | None = None,
    patch_url: Callable[[str], str] | None = None,
) -> xr.Dataset:
    fail_on_error = not skip_broken_datasets

    extra_coords = [] if extra_dims is None else _extract_coords(extra_dims)

    tss = [
        datetime.fromtimestamp(float(ts) * 1e-9)
        for ts in sources.coords["time"].data.ravel()
    ]
    band_query: list[str] = [m.name for m in measurements]
    template = RasterGroupMetadata(
        bands={
            (m.name, 1): RasterBandMetadata(
                m.dtype, m.nodata, m.units, dims=tuple(m.get("dims", ()))
            )
            for m in measurements
        },
        aliases={name: [(name, 1)] for name in band_query},
        extra_dims={coord.dim: len(coord.values) for coord in extra_coords},
        extra_coords=extra_coords,
    )

    load_cfg = {
        m.name: RasterLoadParams(
            m.dtype,
            m.nodata,
            resampling=m.get("resampling", "nearest"),
            fail_on_error=fail_on_error,
            dims=tuple(m.get("dims", ())),
        )
        for m in measurements
    }

    chunks = dask_chunks

    if chunks is not None:
        chunk_shape = resolve_chunk_shape(
            len(tss), geobox, chunks, "float32", cfg=load_cfg
        )
    else:
        chunk_shape = (1, 2048, 2048)

    gbt = GeoboxTiles(geobox, (chunk_shape[1], chunk_shape[2]))

    tyx_bins: dict[tuple[int, int, int], list[int]] = {}
    srcs = []

    if patch_url is None:
        patch_url = lambda x: x  # noqa: E731

    def _dss() -> Generator:
        for tidx, dss in enumerate(sources.data):
            for ds in dss:
                yield tidx, ds

    def _ds_extract(ds: Dataset) -> dict[str, RasterSource]:
        out = {}
        for n in band_query:
            bi = BandInfo(ds, n)
            band_idx = bi.band if bi.band is not None else 1

            if bi.dims is not None and len(bi.dims) > 2:
                band_idx = 0  # 0 indicates extra dims

            out[n] = RasterSource(
                patch_url(bi.uri),
                band=band_idx,
                subdataset=bi.layer,
                geobox=ds_geobox(ds, band=n),
                meta=template.bands[(n, band_idx or 1)],
                driver_data=bi.driver_data,
            )

        return out

    for tidx, ds in _dss():
        srcs.append(_ds_extract(ds))
        for iy, ix in gbt.tiles(ds.extent):
            tyx_bins.setdefault((tidx, iy, ix), []).append(len(srcs) - 1)

    if driver == "kk-debug":
        return SimpleNamespace(  # type: ignore[return-value]
            load_cfg=load_cfg,
            template=template,
            srcs=srcs,
            tyx_bins=tyx_bins,
            gbt=gbt,
            tss=tss,
        )

    rdr = reader_driver(driver)

    return chunked_load(
        load_cfg,
        template,
        srcs,
        tyx_bins=tyx_bins,
        gbt=gbt,
        tss=tss,
        env=rdr.capture_env(),
        rdr=rdr,
        chunks=dask_chunks,
        progress=progress_cbk,
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_index_clone(index_pair_populated_empty) -> None:
    pop_idx, empty_idx = index_pair_populated_empty
    assert list(empty_idx.products.get_all()) == []
    results = empty_idx.clone(pop_idx)
    assert "eo3" in results["metadata_types"].safe
    assert "ga_ls8c_ard_3" in results["products"].safe
    assert results["products"].skipped == 0
    assert results["datasets"].skipped == 0


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_index_clone_small_batch(index_pair_populated_empty) -> None:
    pop_idx, empty_idx = index_pair_populated_empty
    assert list(empty_idx.products.get_all()) == []
    results = empty_idx.clone(pop_idx, batch_size=2)
    assert "eo3" in results["metadata_types"].safe
    assert "ga_ls8c_ard_3" in results["products"].safe
    assert results["products"].skipped == 0
    assert results["datasets"].skipped == 0


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_index_clone_cli(cfg_env_pair, index_pair_populated_empty, clirunner) -> None:
    source_cfg, target_cfg = cfg_env_pair
    clirunner(
        [
            "-E",
            target_cfg._name,
            "system",
            "clone",
            "--lineage-only",
            "--skip-lineage",
            source_cfg._name,
        ],
        skip_env=True,
        expect_success=False,
    )
    clirunner(
        ["-E", target_cfg._name, "system", "clone", source_cfg._name], skip_env=True
    )


@pytest.mark.filterwarnings("ignore::antimeridian.FixWindingWarning")
def test_index_clone_cli_small_batch(
    cfg_env_pair, index_pair_populated_empty, clirunner
) -> None:
    source_cfg, target_cfg = cfg_env_pair
    clirunner(
        [
            "-E",
            target_cfg._name,
            "system",
            "clone",
            "--batch-size",
            "2",
            source_cfg._name,
        ],
        skip_env=True,
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pytest
from odc.geo import xy_

from datacube.testutils import mk_sample_dataset, mk_sample_xr_dataset
from datacube.testutils.io import native_geobox
from datacube.testutils.threads import FakeThreadPoolExecutor


def test_fakethreadpool() -> None:
    def tfunc(a: int, b: int = 0, please_fail=False) -> int:
        if please_fail:
            raise ValueError("as you wish")
        if a == 13:
            raise ValueError("13")
        return a + b

    pool = FakeThreadPoolExecutor()

    assert pool.submit(tfunc, 1).result() == 1
    assert pool.submit(tfunc, 1, 2).result() == 3

    fut = pool.submit(tfunc, 1, please_fail=True)
    assert fut.done()
    assert fut.exception() is not None

    with pytest.raises(ValueError):
        fut.result()

    ff = list(pool.map(tfunc, range(14)))
    assert len(ff) == 14
    assert [f.result() for f in ff[:13]] == list(range(13))
    assert ff[13].exception() is not None

    aa = list(range(10))
    bb = aa[::-1]
    ff = list(pool.map(tfunc, aa, bb))
    assert len(ff) == 10
    assert [f.result() for f in ff[:13]] == [a + b for a, b in zip(aa, bb)]

    pool.shutdown()


def test_mk_sample_xr() -> None:
    xx = mk_sample_xr_dataset()
    assert "band" in xx.data_vars
    assert list(xx.coords) == ["time", "y", "x", "spatial_ref"]
    assert xx.band.dims == ("time", "y", "x")
    assert xx.odc.geobox is not None

    assert mk_sample_xr_dataset(name="xx", shape=(3, 7)).xx.shape == (1, 3, 7)
    assert mk_sample_xr_dataset(name="xx", time=None, shape=(3, 7)).xx.shape == (3, 7)
    assert mk_sample_xr_dataset(name="xx", time=None).xx.dims == ("y", "x")

    assert mk_sample_xr_dataset(resolution=(1, 100)).odc.geobox.resolution == xy_(
        100, 1
    )
    assert mk_sample_xr_dataset(
        resolution=(1, 100), xy=(3, 55)
    ).odc.geobox.transform * (0, 0) == (3, 55)


def test_native_geobox_ingested() -> None:
    from datacube.testutils.geom import AlbersGS
    from datacube.testutils.io import native_geobox

    gbox = AlbersGS.tile_geobox((15, -40))
    ds = mk_sample_dataset(
        [{"name": "a"}], geobox=gbox, product_opts={"with_grid_spec": True}
    )

    assert native_geobox(ds) == gbox

    # check that dataset covering several tiles is detected as invalid
    ds = mk_sample_dataset(
        [{"name": "a"}],
        geobox=gbox.buffered(10, 10),
        product_opts={"with_grid_spec": True},
    )

    with pytest.raises(ValueError):
        native_geobox(ds)


def test_native_geobox_eo3(eo3_dataset_s2) -> None:
    ds = eo3_dataset_s2
    assert ds.crs is not None
    assert "blue" in ds.measurements

    gb1 = native_geobox(ds, basis="blue")

    assert gb1.width == 10980
    assert gb1.height == 10980
    assert gb1.crs == ds.crs

    gb1_ = native_geobox(ds, ("blue", "red", "green"))
    assert gb1_ == gb1

    gb2 = native_geobox(ds, ["swir_1", "swir_2"])
    assert gb2 != gb1
    assert gb2.width == 5490
    assert gb2.height == 5490
    assert gb2.crs == ds.crs

    with pytest.raises(ValueError):
        native_geobox(ds)

    with pytest.raises(ValueError):
        native_geobox(ds, ["no_such_band"])

    ds.metadata_doc["measurements"]["red_edge_1"]["grid"] = "no-such-grid"

    with pytest.raises(ValueError):
        native_geobox(ds, ["red_edge_1"])



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Important functions are:

* :func:`reproject_and_fuse`

"""

import logging
import numbers
from collections import OrderedDict
from collections.abc import Callable, Iterable, Iterator, Sequence
from typing import Any, TypeAlias

import numpy as np
from odc.geo.geobox import GeoBox
from odc.geo.roi import roi_is_empty
from odc.geo.warp import Resampling
from odc.geo.xr import xr_coords
from xarray.core.coordinates import DataArrayCoordinates
from xarray.core.dataarray import DataArray as XrDataArray
from xarray.core.dataset import Dataset as XrDataset

from datacube.drivers._types import ReaderDriver
from datacube.model import Measurement
from datacube.utils import ignore_exceptions_if
from datacube.utils.math import invalid_mask

from ..drivers.datasource import DataSource
from ._base import BandInfo

_LOG: logging.Logger = logging.getLogger(__name__)

FuserFunction: TypeAlias = Callable[[np.ndarray, np.ndarray], Any]  # pylint: disable=invalid-name
ProgressFunction: TypeAlias = Callable[[int, int], Any]  # pylint: disable=invalid-name


def _default_fuser(dst: np.ndarray, src: np.ndarray, dst_nodata) -> None:
    """Overwrite only those pixels in `dst` with `src` that are "not valid"

    For every pixel in dst that equals to dst_nodata replace it with pixel
    from src.
    """
    np.copyto(dst, src, where=invalid_mask(dst, dst_nodata))


def reproject_and_fuse(
    datasources: Sequence[DataSource],
    destination: np.ndarray,
    dst_geobox: GeoBox,
    dst_nodata: int | float | None,
    resampling: Resampling = "nearest",
    fuse_func: FuserFunction | None = None,
    skip_broken_datasets: bool = False,
    progress_cbk: ProgressFunction | None = None,
    extra_dim_index: int | None = None,
) -> np.ndarray:
    """
    Reproject and fuse `sources` into a 2D numpy array `destination`.

    :param datasources: Data sources to open and read from
    :param destination: ndarray of appropriate size to read data into
    :param dst_geobox: GeoBox defining destination region
    :param skip_broken_datasets: Carry on in the face of adversity and failing reads.
    :param progress_cbk: If supplied will be called with 2 integers `Items processed, Total Items`
                         after reading each file.
    """
    # pylint: disable=too-many-locals
    from ._read import read_time_slice

    assert len(destination.shape) == 2

    def copyto_fuser(dest: np.ndarray, src: np.ndarray) -> None:
        _default_fuser(dest, src, dst_nodata)

    fuse_func = fuse_func or copyto_fuser

    destination.fill(dst_nodata)
    if len(datasources) == 0:
        return destination
    elif len(datasources) == 1:
        with ignore_exceptions_if(skip_broken_datasets), datasources[0].open() as rdr:
            read_time_slice(
                rdr, destination, dst_geobox, resampling, dst_nodata, extra_dim_index
            )

        if progress_cbk:
            progress_cbk(1, 1)

        return destination
    else:
        # Multiple sources, we need to fuse them together into a single array
        buffer_ = np.full(destination.shape, dst_nodata, dtype=destination.dtype)
        for n_so_far, source in enumerate(datasources, 1):
            with ignore_exceptions_if(skip_broken_datasets):
                with source.open() as rdr:
                    roi = read_time_slice(
                        rdr,
                        buffer_,
                        dst_geobox,
                        resampling,
                        dst_nodata,
                        extra_dim_index,
                    )

                if not roi_is_empty(roi):
                    fuse_func(destination[roi], buffer_[roi])
                    if dst_nodata is not None:
                        buffer_[roi] = dst_nodata  # clean up for next read
                    elif issubclass(destination.dtype.type, numbers.Real):
                        buffer_[roi] = float("nan")
                    else:
                        buffer_[roi] = 0

            if progress_cbk:
                progress_cbk(n_so_far, len(datasources))

        return destination


def _mk_empty_ds(coords: DataArrayCoordinates, geobox: GeoBox) -> XrDataset:
    cc = OrderedDict(coords.items())
    cc.update(xr_coords(geobox, None))
    return XrDataset(coords=cc, attrs={"crs": geobox.crs})


def _allocate_storage(
    coords: DataArrayCoordinates, geobox: GeoBox, measurements: Iterable[Measurement]
) -> XrDataset:
    xx = _mk_empty_ds(coords, geobox)
    dims = list(xx.coords.keys())
    shape = tuple(xx.sizes[k] for k in dims)

    for m in measurements:
        name, dtype, attrs = m.name, m.dtype, m.dataarray_attrs()
        attrs["crs"] = geobox.crs
        data = np.empty(shape, dtype=dtype)
        xx[name] = XrDataArray(
            data, coords=xx.coords, dims=dims, name=name, attrs=attrs
        )

    return xx


def xr_load(
    sources: XrDataArray,
    geobox: GeoBox,
    measurements: list[Measurement],
    driver: ReaderDriver,
    driver_ctx_prev: Any | None = None,
    skip_broken_datasets: bool = False,
) -> tuple[XrDataset, Any]:
    # pylint: disable=too-many-locals
    from ._read import read_time_slice_v2

    out = _allocate_storage(sources.coords, geobox, measurements)

    def all_groups() -> Iterator[tuple[Measurement, tuple[int, ...], list[BandInfo]]]:
        for idx, dss in np.ndenumerate(sources.values):
            for m in measurements:
                bbi = [BandInfo(ds, m.name) for ds in dss]
                yield m, idx, bbi

    def just_bands(groups) -> Iterator[BandInfo]:
        for _, _, bbi in groups:
            yield from bbi

    groups = list(all_groups())
    ctx = driver.new_load_context(just_bands(groups), driver_ctx_prev)

    # TODO: run upto N concurrently
    for m, idx, bbi in groups:
        dst = out.data_vars[m.name].values[idx]
        dst[:] = m.nodata
        resampling = m.get("resampling_method", "nearest")
        fuse_func = m.get("fuser", None)

        for band in bbi:
            rdr = driver.open(band, ctx).result()

            pix, roi = read_time_slice_v2(rdr, geobox, resampling, m.nodata)

            if pix is not None:
                if fuse_func:
                    fuse_func(dst[roi], pix)
                else:
                    _default_fuser(dst[roi], pix, m.nodata)

    return out, ctx



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import string
from uuid import uuid4

import netCDF4
import numpy
import pytest
import xarray as xr
from hypothesis import given
from hypothesis.strategies import text
from odc.geo import CRS
from odc.geo.geobox import Coordinate

from datacube.drivers.netcdf import write_dataset_to_netcdf
from datacube.drivers.netcdf._write import _get_units
from datacube.drivers.netcdf.writer import (
    DEFAULT_GRID_MAPPING,
    Variable,
    create_coordinate,
    create_grid_mapping_variable,
    create_netcdf,
    create_variable,
    flag_mask_meanings,
    netcdfy_data,
)
from datacube.testutils import mk_sample_xr_dataset
from datacube.utils import DatacubeException, read_strings_from_netcdf

crs_var = DEFAULT_GRID_MAPPING

GEO_PROJ = CRS("""GEOGCS["WGS 84",
                           DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],
                           AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433],
                           AUTHORITY["EPSG","4326"]]""")

ALBERS_PROJ = CRS("""PROJCS["GDA94 / Australian Albers",
                                GEOGCS["GDA94",
                                    DATUM["Geocentric_Datum_of_Australia_1994",
                                        SPHEROID["GRS 1980",6378137,298.257222101,
                                            AUTHORITY["EPSG","7019"]],
                                        TOWGS84[0,0,0,0,0,0,0],
                                        AUTHORITY["EPSG","6283"]],
                                    PRIMEM["Greenwich",0,
                                        AUTHORITY["EPSG","8901"]],
                                    UNIT["degree",0.01745329251994328,
                                        AUTHORITY["EPSG","9122"]],
                                    AUTHORITY["EPSG","4283"]],
                                UNIT["metre",1,
                                    AUTHORITY["EPSG","9001"]],
                                PROJECTION["Albers_Conic_Equal_Area"],
                                PARAMETER["standard_parallel_1",-18],
                                PARAMETER["standard_parallel_2",-36],
                                PARAMETER["latitude_of_center",0],
                                PARAMETER["longitude_of_center",132],
                                PARAMETER["false_easting",0],
                                PARAMETER["false_northing",0],
                                AUTHORITY["EPSG","3577"],
                                AXIS["Easting",EAST],
                                AXIS["Northing",NORTH]]""")

SINIS_PROJ = CRS("""PROJCS["Sinusoidal",
                                GEOGCS["GCS_Undefined",
                                        DATUM["Undefined",
                                        SPHEROID["User_Defined_Spheroid",6371007.181,0.0]],
                                        PRIMEM["Greenwich",0.0],
                                        UNIT["Degree",0.0174532925199433]],
                                PROJECTION["Sinusoidal"],
                                PARAMETER["False_Easting",0.0],
                                PARAMETER["False_Northing",0.0],
                                PARAMETER["Central_Meridian",0.0],
                                UNIT["Meter",1.0]]""")

LCC2_PROJ = CRS("""PROJCS["unnamed",
                               GEOGCS["WGS 84",
                                       DATUM["unknown",
                                       SPHEROID["WGS84",6378137,6556752.3141]],
                                       PRIMEM["Greenwich",0],
                                       UNIT["degree",0.0174532925199433]],
                               UNIT["metre",1, AUTHORITY["EPSG","9001"]],
                               PROJECTION["Lambert_Conformal_Conic_2SP"],
                               PARAMETER["standard_parallel_1",17.5],
                               PARAMETER["standard_parallel_2",29.5],
                               PARAMETER["latitude_of_origin",12],
                               PARAMETER["central_meridian",-102],
                               PARAMETER["false_easting",2500000],
                               PARAMETER["false_northing",0]]""")


def _ensure_spheroid(var) -> None:
    assert "semi_major_axis" in var.ncattrs()
    assert "semi_minor_axis" in var.ncattrs()
    assert "inverse_flattening" in var.ncattrs()


def _ensure_gdal(var) -> None:
    assert "GeoTransform" in var.ncattrs()
    assert "spatial_ref" in var.ncattrs()


def _ensure_geospatial(nco) -> None:
    assert "geospatial_bounds" in nco.ncattrs()
    assert "geospatial_bounds_crs" in nco.ncattrs()
    assert nco.getncattr("geospatial_bounds_crs") == "EPSG:4326"

    assert "geospatial_lat_min" in nco.ncattrs()
    assert "geospatial_lat_max" in nco.ncattrs()
    assert "geospatial_lat_units" in nco.ncattrs()
    assert nco.getncattr("geospatial_lat_units") == "degrees_north"

    assert "geospatial_lon_min" in nco.ncattrs()
    assert "geospatial_lon_max" in nco.ncattrs()
    assert "geospatial_lon_units" in nco.ncattrs()
    assert nco.getncattr("geospatial_lon_units") == "degrees_east"


def test_create_albers_projection_netcdf(tmpnetcdf_filename) -> None:
    nco = create_netcdf(tmpnetcdf_filename)
    create_coordinate(nco, "x", numpy.array([1.0, 2.0, 3.0]), "m")
    create_coordinate(nco, "y", numpy.array([1.0, 2.0, 3.0]), "m")
    create_grid_mapping_variable(nco, ALBERS_PROJ)
    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert crs_var in nco.variables
        assert nco[crs_var].grid_mapping_name == "albers_conical_equal_area"
        assert "standard_parallel" in nco[crs_var].ncattrs()
        assert "longitude_of_central_meridian" in nco[crs_var].ncattrs()
        assert "latitude_of_projection_origin" in nco[crs_var].ncattrs()
        _ensure_spheroid(nco[crs_var])
        _ensure_gdal(nco[crs_var])
        _ensure_geospatial(nco)


def test_create_lambert_conformal_conic_2sp_projection_netcdf(
    tmpnetcdf_filename,
) -> None:
    nco = create_netcdf(tmpnetcdf_filename)
    create_coordinate(nco, "x", numpy.array([1.0, 2.0, 3.0]), "m")
    create_coordinate(nco, "y", numpy.array([1.0, 2.0, 3.0]), "m")
    create_grid_mapping_variable(nco, LCC2_PROJ)
    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert crs_var in nco.variables
        assert nco[crs_var].grid_mapping_name == "lambert_conformal_conic"
        assert "standard_parallel" in nco[crs_var].ncattrs()
        assert "longitude_of_central_meridian" in nco[crs_var].ncattrs()
        assert "latitude_of_projection_origin" in nco[crs_var].ncattrs()
        assert "false_easting" in nco[crs_var].ncattrs()
        assert "false_northing" in nco[crs_var].ncattrs()
        _ensure_spheroid(nco[crs_var])
        _ensure_gdal(nco[crs_var])
        _ensure_geospatial(nco)


def test_create_epsg4326_netcdf(tmpnetcdf_filename) -> None:
    nco = create_netcdf(tmpnetcdf_filename)
    create_coordinate(nco, "latitude", numpy.array([1.0, 2.0, 3.0]), "m")
    create_coordinate(nco, "longitude", numpy.array([1.0, 2.0, 3.0]), "m")
    create_grid_mapping_variable(nco, GEO_PROJ)
    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert crs_var in nco.variables
        assert nco[crs_var].grid_mapping_name == "latitude_longitude"
        _ensure_spheroid(nco[crs_var])
        _ensure_geospatial(nco)


def test_create_sinus_netcdf(tmpnetcdf_filename) -> None:
    nco = create_netcdf(tmpnetcdf_filename)
    create_coordinate(nco, "x", numpy.array([1.0, 2.0, 3.0]), "m")
    create_coordinate(nco, "y", numpy.array([1.0, 2.0, 3.0]), "m")
    create_grid_mapping_variable(nco, SINIS_PROJ)
    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert crs_var in nco.variables
        assert nco[crs_var].grid_mapping_name == "sinusoidal"
        assert "longitude_of_central_meridian" in nco[crs_var].ncattrs()
        _ensure_spheroid(nco[crs_var])
        _ensure_geospatial(nco)


# Work around outstanding bug with hypothesis/pytest, where function level fixtures are only run once.
# Generate a new netcdf filename for each run, so that old files don't cause permission errors on windows
# due to antivirus software filesystem lag.
# See https://github.com/HypothesisWorks/hypothesis-python/issues/377
@given(
    s1=text(alphabet=string.printable, max_size=100),
    s2=text(alphabet=string.printable, max_size=100),
    s3=text(alphabet=string.printable, max_size=100),
)
def test_create_string_variable(workdir, s1, s2, s3) -> None:
    tmpnetcdf_filename = workdir.join(f"testfile_np_{uuid4()}.nc")
    str_var = "str_var"
    nco = create_netcdf(tmpnetcdf_filename)
    coord = create_coordinate(nco, "greg", numpy.array([1.0, 3.0, 9.0]), "cubic gregs")
    assert coord is not None

    dtype = numpy.dtype("S100")
    data = numpy.array([s1, s2, s3], dtype=dtype)

    var = create_variable(nco, str_var, Variable(dtype, None, ("greg",), None))
    var[:] = netcdfy_data(data)
    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert str_var in nco.variables

    for returned, expected in zip(
        read_strings_from_netcdf(tmpnetcdf_filename, variable=str_var), (s1, s2, s3)
    ):
        assert returned == expected


def test_chunksizes(tmpnetcdf_filename) -> None:
    nco = create_netcdf(tmpnetcdf_filename)

    x = numpy.arange(3, dtype="float32")
    y = numpy.arange(5, dtype="float32")

    coord1 = create_coordinate(nco, "x", x, "m")
    coord2 = create_coordinate(nco, "y", y, "m")

    assert coord1 is not None and coord2 is not None

    no_chunks = create_variable(
        nco, "no_chunks", Variable(numpy.dtype("int16"), None, ("x", "y"), None)
    )

    min_max_chunks = create_variable(
        nco,
        "min_max_chunks",
        Variable(numpy.dtype("int16"), None, ("x", "y"), None),
        chunksizes=(2, 50),
    )

    assert no_chunks is not None
    assert min_max_chunks is not None

    strings = numpy.array(["AAa", "bbb", "CcC"], dtype="S")
    strings = xr.DataArray(strings, dims=["x"], coords={"x": x})
    create_variable(nco, "strings_unchunked", strings)
    create_variable(nco, "strings_chunked", strings, chunksizes=(1,))

    nco.close()

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        assert nco["no_chunks"].chunking() == "contiguous"
        assert nco["min_max_chunks"].chunking() == [2, 5]
        assert nco["strings_unchunked"].chunking() == "contiguous"
        assert nco["strings_chunked"].chunking() == [1, 3]


EXAMPLE_FLAGS_DEF = {
    "band_1_saturated": {
        "bits": 0,
        "values": {0: True, 1: False},
        "description": "Band 1 is saturated",
    },
    "band_2_saturated": {
        "bits": 1,
        "values": {0: True, 1: False},
        "description": "Band 2 is saturated",
    },
    "band_3_saturated": {
        "bits": 2,
        "values": {0: True, 1: False},
        "description": "Band 3 is saturated",
    },
    "land_sea": {
        "bits": 9,
        "values": {0: "sea", 1: "land"},
        "description": "Land/Sea observation",
    },
}


def test_measurements_model_netcdfflags() -> None:
    masks, valid_range, meanings = flag_mask_meanings(EXAMPLE_FLAGS_DEF)
    assert (valid_range == numpy.array([0, 1023])).all()
    assert (masks == numpy.array([1, 2, 4, 512])).all()
    assert (
        meanings == "no_band_1_saturated no_band_2_saturated no_band_3_saturated land"
    )


def test_useful_error_on_write_empty_dataset(tmpnetcdf_filename) -> None:
    with pytest.raises(DatacubeException) as excinfo:
        ds = xr.Dataset()
        write_dataset_to_netcdf(ds, tmpnetcdf_filename)
    assert "empty" in str(excinfo.value)

    with pytest.raises(DatacubeException) as excinfo:
        ds = xr.Dataset(data_vars={"blue": (("time",), numpy.array([0, 1, 2]))})
        write_dataset_to_netcdf(ds, tmpnetcdf_filename)
    assert "geobox" in str(excinfo.value)


@pytest.mark.parametrize("odc_style_xr_dataset", [None], indirect=True)
def test_write_dataset_to_netcdf(tmpnetcdf_filename, odc_style_xr_dataset) -> None:
    write_dataset_to_netcdf(
        odc_style_xr_dataset,
        tmpnetcdf_filename,
        global_attributes={"foo": "bar"},
        variable_params={"B10": {"attrs": {"abc": "xyz"}}},
    )

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        nco.set_auto_mask(False)
        assert "B10" in nco.variables
        var = nco.variables["B10"]
        assert (var[:] == odc_style_xr_dataset["B10"].values).all()

        assert "foo" in nco.ncattrs()
        assert nco.getncattr("foo") == "bar"

        assert "abc" in var.ncattrs()
        assert var.getncattr("abc") == "xyz"

    with pytest.raises(RuntimeError):
        write_dataset_to_netcdf(odc_style_xr_dataset, tmpnetcdf_filename)

    # Check grid_mapping is a coordinate
    xx = xr.open_dataset(tmpnetcdf_filename)
    assert crs_var in xx.coords
    assert crs_var not in xx.data_vars


def test_write_dataset_with_time_dimension_to_netcdf(tmpnetcdf_filename) -> None:
    xx = mk_sample_xr_dataset(name="B10", time="2020-01-01")
    assert "time" in xx.coords
    assert "units" not in xx.time.attrs

    write_dataset_to_netcdf(
        xx,
        tmpnetcdf_filename,
        global_attributes={"foo": "bar"},
        variable_params={"B10": {"attrs": {"abc": "xyz"}}},
    )

    with netCDF4.Dataset(tmpnetcdf_filename) as nco:
        nco.set_auto_mask(False)
        assert "B10" in nco.variables
        var = nco.variables["B10"]
        assert (var[:] == xx["B10"].values).all()

        assert "foo" in nco.ncattrs()
        assert nco.getncattr("foo") == "bar"

        assert "abc" in var.ncattrs()
        assert var.getncattr("abc") == "xyz"

    with pytest.raises(RuntimeError):
        write_dataset_to_netcdf(xx, tmpnetcdf_filename)

    # Check grid_mapping is a coordinate
    yy = xr.open_dataset(tmpnetcdf_filename)
    assert crs_var in yy.coords
    assert crs_var not in yy.data_vars

    assert "time" in yy.coords


def test_get_units() -> None:
    assert _get_units(Coordinate([], "K", None)) == "K"
    assert _get_units(Coordinate([], None, None)) == "1"
    assert _get_units(Coordinate(numpy.zeros(1, dtype="uint8"), None, None)) == "1"
    assert _get_units(
        Coordinate(numpy.zeros(1, dtype="datetime64[s]"), None, None)
    ).startswith("seconds since ")



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
User configuration.
"""

import warnings

from datacube.migration import ODC2DeprecationWarning

warnings.warn(
    "The old datacube.config  is no longer supported.  Please use the new datacube.cfg library",
    ODC2DeprecationWarning,
)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
from collections import namedtuple
from contextlib import contextmanager
from copy import deepcopy
from typing import cast
from uuid import UUID

from datacube.drivers.postgres import PostgresDb
from datacube.index.exceptions import DuplicateRecordError
from datacube.index.postgres._datasets import DatasetResource
from datacube.index.postgres.index import Index
from datacube.model import Dataset, MetadataType, Product

_nbar_uuid = UUID("f2f12372-8366-11e5-817e-1040f381a756")
_ortho_uuid = UUID("5cf41d98-eda9-11e4-8a8e-1040f381a756")
_telemetry_uuid = UUID("4ec8fe97-e8b9-11e4-87ff-1040f381a756")

# An NBAR with source datasets. Many fields have been removed to keep it semi-focused to our ingest test.
_EXAMPLE_NBAR = {
    "id": str(_nbar_uuid),
    "product_type": "nbar_brdf",
    "checksum_path": "package.sha1",
    "ga_label": "LS8_OLITIRS_NBAR_P54_GALPGS01-002_112_079_20140126",
    "ga_level": "P54",
    "size_bytes": 4550,
    "platform": {"code": "LANDSAT_8"},
    "creation_dt": datetime.datetime(2014, 1, 26, 2, 5, 23, 126373).isoformat(),
    "instrument": {"name": "OLI_TIRS"},
    "format": {"name": "GeoTIFF"},
    "extent": {
        "center_dt": datetime.datetime(2014, 1, 26, 2, 5, 23, 126373).isoformat(),
        "coord": {
            "ul": {"lat": -26.37259, "lon": 116.58914},
            "lr": {"lat": -28.48062, "lon": 118.96145},
            "ur": {"lat": -26.36025, "lon": 118.92432},
            "ll": {"lat": -28.49412, "lon": 116.58121},
        },
    },
    "lineage": {
        "machine": {},
        "source_datasets": {
            "ortho": {
                "product_level": "L1T",
                "product_type": "ortho",
                "id": str(_ortho_uuid),
                "usgs": {"scene_id": "LC81120792014026ASA00"},
                "extent": {
                    "center_dt": datetime.datetime(
                        2014, 1, 26, 2, 5, 23, 126373
                    ).isoformat(),
                    "coord": {
                        "ul": {"lat": -26.37259, "lon": 116.58914},
                        "lr": {"lat": -28.48062, "lon": 118.96145},
                        "ur": {"lat": -26.36025, "lon": 118.92432},
                        "ll": {"lat": -28.49412, "lon": 116.58121},
                    },
                },
                "size_bytes": 1854924494,
                "platform": {"code": "LANDSAT_8"},
                "creation_dt": datetime.datetime(2015, 4, 7, 0, 58, 8).isoformat(),
                "instrument": {"name": "OLI_TIRS"},
                "checksum_path": "package.sha1",
                "ga_label": "LS8_OLITIRS_OTH_P51_GALPGS01-002_112_079_20140126",
                "grid_spatial": {
                    "projection": {
                        "spatial_reference": "EPSG:28350",
                        "resampling_option": "CUBIC_CONVOLUTION",
                        "geo_ref_points": {
                            "ul": {"y": 7082987.5, "x": 459012.5},
                            "lr": {"y": 6847987.5, "x": 692012.5},
                            "ur": {"y": 7082987.5, "x": 692012.5},
                            "ll": {"y": 6847987.5, "x": 459012.5},
                        },
                        "orientation": "NORTH_UP",
                    }
                },
                "acquisition": {
                    "groundstation": {
                        "code": "ASA",
                        "eods_domain_code": "002",
                        "label": "Alice Springs",
                    }
                },
                "format": {"name": "GEOTIFF"},
                "lineage": {
                    "algorithm": {"name": "LPGS", "parameters": {}, "version": "2.4.0"},
                    "machine": {},
                    "source_datasets": {
                        "satellite_telemetry_data": {
                            "product_type": "satellite_telemetry_data",
                            "checksum_path": "package.sha1",
                            "id": str(_telemetry_uuid),
                            "ga_label": "LS8_OLITIRS_STD-MD_P00_LC81160740742015089ASA00_"
                            "116_074_20150330T022553Z20150330T022657",
                            "ga_level": "P00",
                            "size_bytes": 637660782,
                            "platform": {"code": "LANDSAT_8"},
                            "creation_dt": datetime.datetime(
                                2015, 4, 22, 6, 32, 4
                            ).isoformat(),
                            "instrument": {"name": "OLI_TIRS"},
                            "format": {"name": "MD"},
                            "lineage": {"source_datasets": {}},
                        }
                    },
                },
            }
        },
    },
}

_EXAMPLE_METADATA_TYPE = MetadataType(
    {
        "name": "eo",
        "dataset": {
            "id": ["id"],
            "label": ["ga_label"],
            "creation_time": ["creation_dt"],
            "measurements": ["image", "bands"],
            "sources": ["lineage", "source_datasets"],
        },
    }
)

_EXAMPLE_PRODUCT = Product(
    _EXAMPLE_METADATA_TYPE,
    {"name": "eo", "description": "", "metadata_type": "eo", "metadata": {}},
)


def _build_dataset(doc):
    sources = {
        name: _build_dataset(src)
        for name, src in doc["lineage"]["source_datasets"].items()
    }
    return Dataset(_EXAMPLE_PRODUCT, doc, uri="file://test.zzz", sources=sources)


_EXAMPLE_NBAR_DATASET = _build_dataset(_EXAMPLE_NBAR)

DatasetRecord = namedtuple(
    "DatasetRecord",
    ["id", "metadata", "dataset_type_ref", "uris", "added", "added_by", "archived"],
)


class MockDb:
    def __init__(self) -> None:
        self.dataset: dict = {}
        self.dataset_source: set = set()

    @contextmanager
    def _connect(self):
        yield self

    def begin(self) -> None:
        pass

    def commit(self) -> None:
        pass

    def rollback(self) -> None:
        pass

    def get_dataset(self, id_):
        return self.dataset.get(id_, None)

    def get_locations(self, dataset):
        return ["file:xxx"]

    def datasets_intersection(self, ids):
        return [k for k in ids if k in self.dataset]

    def contains_dataset(self, dataset_id):
        return dataset_id in self.dataset

    def insert_dataset_location(self, *args, **kwargs) -> None:
        return

    def insert_dataset(self, metadata_doc, dataset_id, dataset_type_id) -> bool:
        # Will we pretend this one was already ingested?
        if dataset_id in self.dataset:
            raise DuplicateRecordError("already ingested")

        self.dataset[dataset_id] = DatasetRecord(
            dataset_id, deepcopy(metadata_doc), dataset_type_id, None, None, None, None
        )
        return True

    def insert_dataset_source(self, classifier, dataset_id, source_dataset_id) -> None:
        self.dataset_source.add((classifier, dataset_id, source_dataset_id))


class MockTypesResource:
    def __init__(self, type_) -> None:
        self.type = type_

    def get(self, *args, **kwargs):
        return self.type

    def get_by_name(self, *args, **kwargs):
        return self.type

    def get_by_name_unsafe(self, *args, **kwargs):
        return self.type

    @contextmanager
    def _db_connection(self, transaction=False):
        yield MockDb()


class MockIndex:
    def __init__(self, db, product) -> None:
        self._db = db
        self.products = MockTypesResource(product)

    def thread_transaction(self):
        return None

    @contextmanager
    def _active_connection(self, transaction=False):
        yield self._db


def test_index_dataset() -> None:
    mock_db = MockDb()
    mock_index = MockIndex(mock_db, _EXAMPLE_PRODUCT)
    datasets = DatasetResource(cast(PostgresDb, mock_db), cast(Index, mock_index))
    datasets.add(_EXAMPLE_NBAR_DATASET)

    ids = {d.id for d in mock_db.dataset.values()}
    assert ids == {_nbar_uuid, _ortho_uuid, _telemetry_uuid}

    # Three datasets (ours and the two embedded source datasets)
    assert len(mock_db.dataset) == 3

    # Our three datasets should be linked together
    # Nbar -> Ortho -> Telemetry
    assert len(mock_db.dataset_source) == 2
    assert mock_db.dataset_source == {
        ("ortho", _nbar_uuid, _ortho_uuid),
        ("satellite_telemetry_data", _ortho_uuid, _telemetry_uuid),
    }

    # Nothing ingested, because we reported the first as already ingested.
    datasets.add(_EXAMPLE_NBAR_DATASET)
    assert len(mock_db.dataset) == 3
    assert len(mock_db.dataset_source) == 2


def test_index_already_ingested_source_dataset() -> None:
    mock_db = MockDb()
    mock_index = MockIndex(mock_db, _EXAMPLE_PRODUCT)
    datasets = DatasetResource(cast(PostgresDb, mock_db), cast(Index, mock_index))
    datasets.add(_EXAMPLE_NBAR_DATASET.sources["ortho"])

    assert len(mock_db.dataset) == 2
    assert len(mock_db.dataset_source) == 1

    datasets.add(_EXAMPLE_NBAR_DATASET)
    assert len(mock_db.dataset) == 3
    assert len(mock_db.dataset_source) == 2


def test_index_two_levels_already_ingested() -> None:
    mock_db = MockDb()
    mock_index = MockIndex(mock_db, _EXAMPLE_PRODUCT)
    datasets = DatasetResource(cast(PostgresDb, mock_db), cast(Index, mock_index))
    datasets.add(
        _EXAMPLE_NBAR_DATASET.sources["ortho"].sources["satellite_telemetry_data"]
    )

    assert len(mock_db.dataset) == 1
    assert len(mock_db.dataset_source) == 0

    datasets.add(_EXAMPLE_NBAR_DATASET)
    assert len(mock_db.dataset) == 3
    assert len(mock_db.dataset_source) == 2



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

from ._tools import singleton_setup
from .driver_cache import load_drivers


class WriterDriverCache:
    def __init__(self, group: str) -> None:
        self._drivers = load_drivers(group)

        for driver in list(self._drivers.values()):
            if hasattr(driver, "aliases"):
                for alias in driver.aliases:
                    self._drivers[alias] = driver

    def __call__(self, name: str) -> dict | None:
        """
        :returns: None if driver with a given name is not found

        :param name: Driver name
        :return: Returns WriterDriver
        """
        return self._drivers.get(name, None)

    def drivers(self) -> list[str]:
        """
        Returns list of driver names
        """
        return list(self._drivers.keys())


def writer_cache() -> WriterDriverCache:
    """
    Singleton for WriterDriverCache
    """
    return singleton_setup(
        writer_cache, "_instance", WriterDriverCache, "datacube.plugins.io.write"
    )


def writer_drivers() -> list[str]:
    """
    Returns list driver names
    """
    return writer_cache().drivers()


def storage_writer_by_name(name: str):
    """
    Lookup writer driver by name

    :return: Initialised writer driver instance or None if driver with this name doesn't exist
    """
    return writer_cache()(name)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Catalog of virtual products.
"""

from collections.abc import Iterable, Iterator, Mapping
from itertools import chain
from typing import Any

import yaml
from typing_extensions import override

from datacube.model.utils import SafeDumper


class UnappliedTransform:
    def __init__(self, name_resolver, recipe) -> None:
        self.name_resolver = name_resolver
        self.recipe = recipe

    def __call__(self, input) -> Any:  # noqa: A002
        return self.name_resolver.construct(**self.recipe, input=input)

    @override
    def __repr__(self) -> str:
        return yaml.dump(
            self.recipe, Dumper=SafeDumper, default_flow_style=False, indent=2
        )


class Catalog(Mapping):
    """
    A catalog of virtual products specified in a yaml document.
    """

    def __init__(self, name_resolver, contents) -> None:
        self.name_resolver = name_resolver
        self.contents = contents
        common = set(self._names("products")) & set(self._names("transforms"))
        assert not common, f"common names found in products and transforms {common}"

    def _names(self, section) -> list:
        """List of names under a section (products or transforms)."""
        if section not in self.contents:
            return []
        return list(self.contents[section])

    @override
    def __getitem__(self, name: str):
        """
        Looks up a virtual product or transform by name.
        Returns `None` if not found.
        """
        if name in self._names("products"):
            return self.name_resolver.construct(
                **self.contents["products"][name]["recipe"]
            )
        if name in self._names("transforms"):
            return UnappliedTransform(
                self.name_resolver, self.contents["transforms"][name]["recipe"]
            )

        # raising a `KeyError` here stops autocompletion from working
        return None

    def __getattr__(self, name: str):
        return self[name]

    @override
    def __len__(self) -> int:
        return len(self._names("products")) + len(self._names("transforms"))

    @override
    def __iter__(self) -> Iterator:
        return chain(iter(self._names("products")), iter(self._names("transforms")))

    @override
    def __dir__(self) -> Iterable[str]:
        """
        Override to provide autocompletion of products and transforms.
        """
        return sorted(
            dir(Mapping)
            + list(self.__dict__)
            + self._names("products")
            + self._names("transforms")
        )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Build and index fields within documents.
"""

import math
from collections import namedtuple
from collections.abc import Callable, Sequence
from datetime import date, datetime, timezone
from decimal import Decimal
from typing import Any, TypeAlias

from sqlalchemy import and_, cast, func
from sqlalchemy.dialects import postgresql as postgres
from sqlalchemy.dialects.postgresql import INTERVAL, NUMRANGE, TSTZRANGE
from sqlalchemy.dialects.postgresql import Range as PgRange
from sqlalchemy.orm import aliased
from sqlalchemy.sql import ColumnElement, ColumnExpressionArgument, FromClause
from sqlalchemy.types import TIMESTAMP
from typing_extensions import override

from datacube import utils
from datacube.drivers.postgis._schema import Dataset, search_field_index_map
from datacube.model import Range
from datacube.model.fields import Expression, Field
from datacube.utils import cached_property, get_doc_offset
from datacube.utils.changes import Offset
from datacube.utils.dates import tz_as_utc

DatasetJoinArgs: TypeAlias = (
    tuple[FromClause] | tuple[FromClause, ColumnExpressionArgument]
)


class PgField(Field):
    """
    Postgis implementation of a searchable field. May be a value inside
    a JSONB column.
    """

    def __init__(
        self,
        name: str,
        description: str,
        alchemy_column: ColumnElement,
        indexed: bool,
        alchemy_table: FromClause | None = None,
    ) -> None:
        super().__init__(name, description)

        # The underlying SQLAlchemy column. (eg. DATASET.c.metadata)
        self.alchemy_column = alchemy_column
        self.indexed = indexed
        self.alchemy_table = alchemy_table

    @property
    def select_alchemy_table(self) -> FromClause:
        if self.alchemy_table is not None:
            return self.alchemy_table
        return self.alchemy_column.table

    def normalise_value(self, value):
        """
        Wrap the given value with any necessary type casts/conversions for this field.

        Overridden by other classes as needed.
        """
        # Default do nothing (eg. string datatypes)
        return value

    @cached_property
    def search_index_table(self) -> FromClause:
        if self.indexed:
            search_table = search_field_index_map[self.type_name]
            return aliased(  # type: ignore[return-value]
                search_table, name=f"{search_table.__tablename__}-{self.name}"
            )
        else:
            return self.select_alchemy_table

    @cached_property
    def dataset_join_args(self) -> DatasetJoinArgs:
        if self.indexed:
            return (
                self.search_index_table,
                and_(
                    Dataset.id == self.search_index_table.dataset_ref,
                    self.search_index_table.search_key == self.name,
                ),
            )
        else:
            return (self.search_index_table,)

    @property
    def alchemy_expression(self) -> ColumnExpressionArgument:
        """
        Get an SQLAlchemy expression for accessing this field.
        :return:
        """
        raise NotImplementedError("alchemy expression")

    @property
    def search_alchemy_expression(self) -> ColumnExpressionArgument:
        if self.indexed:
            return self.search_index_table.search_val
        else:
            return self.alchemy_expression

    @property
    def sql_expression(self):
        """
        Get the raw SQL expression for this field as a string.
        :rtype: str
        """
        return str(
            self.alchemy_expression.compile(
                dialect=postgres.dialect(), compile_kwargs={"literal_binds": True}
            )
        )

    @override
    def __eq__(self, value) -> Expression:  # type: ignore[override]
        return EqualsExpression(self, value)

    @override
    def between(self, low, high) -> Expression:
        raise NotImplementedError("between expression")


class NativeField(PgField):
    """
    Fields hard-coded into the schema. (not user configurable)
    """

    def __init__(
        self,
        name: str,
        description: str,
        alchemy_column: ColumnElement,
        alchemy_expression: ColumnExpressionArgument | None = None,
        join_clause: ColumnExpressionArgument | None = None,
        alchemy_table: FromClause | None = None,
        # Should this be selected by default when selecting all fields?
        affects_row_selection: bool = False,
    ) -> None:
        super().__init__(
            name,
            description,
            alchemy_column,
            indexed=False,
            alchemy_table=alchemy_table,
        )
        self._expression = alchemy_expression
        self.affects_row_selection = affects_row_selection
        self.join_clause = join_clause

    @property
    @override
    def alchemy_expression(self) -> ColumnExpressionArgument:
        expression = (
            self._expression if self._expression is not None else self.alchemy_column
        )
        return expression.label(self.name)  # type: ignore[union-attr]

    @override
    @cached_property
    def dataset_join_args(self) -> DatasetJoinArgs:
        if self.join_clause is None:
            return super().dataset_join_args
        return self.select_alchemy_table, self.join_clause


class PgDocField(PgField):
    """
    A field extracted from inside a (jsonb) document.
    """

    def value_to_alchemy(self, value):
        """
        Wrap the given value with any necessary type casts/conversions for this field.

        Overridden by other classes as needed.
        """
        # Default do nothing (eg. string datatypes)
        return value

    def search_value_to_alchemy(self, value):
        """
        Value to use in search tables. Identical to value_to_alchemy, unless it needs
        to be promoted to range type.
        """
        return self.value_to_alchemy(value)

    def parse_value(self, value):
        """
        Parse the value from a string. May be overridden by subclasses.
        """
        return value

    def _alchemy_offset_value(
        self,
        doc_offsets: Sequence[Offset],
        agg_function: Callable[[Any], ColumnElement],
        type_: type | None = None,
    ) -> ColumnElement:
        """
        Get an sqlalchemy value for the given offsets of this field's sqlalchemy column.
        If there are multiple they will be combined using the given aggregate function.

        Offsets can either be single:
            ('platform', 'code')
        Or multiple:
            (('platform', 'code'), ('satellite', 'name'))

        In the latter case, the multiple values are combined using the given aggregate function
        (defaults to using coalesce: grab the first non-null value)
        """
        if not doc_offsets:
            raise ValueError("Value requires at least one offset")

        if isinstance(doc_offsets[0], str):
            # It's a single offset.
            doc_offsets = [doc_offsets]

        alchemy_values = [
            self.value_to_alchemy(self.alchemy_column[offset].astext)
            for offset in doc_offsets
        ]
        if type_ is not None:
            alchemy_values = [cast(v, type_) for v in alchemy_values]
        # If there's multiple fields, we aggregate them (eg. "min()"). Otherwise use the one.
        return (
            agg_function(*alchemy_values)
            if len(alchemy_values) > 1
            else alchemy_values[0]
        )

    def _extract_offset_value(self, doc, doc_offsets, agg_function):
        """
        Extract a value for the given document offsets.

        Same as _alchemy_offset_value(), but returns the value instead of an sqlalchemy expression to calc the value.
        """
        if not doc_offsets:
            raise ValueError("Value requires at least one offset")

        if isinstance(doc_offsets[0], str):
            # It's a single offset.
            doc_offsets = [doc_offsets]

        values = (get_doc_offset(offset, doc) for offset in doc_offsets)
        values = [self.parse_value(v) for v in values if v is not None]

        if not values:
            return None
        if len(values) == 1:
            return values[0]
        return agg_function(*values)


class SimpleDocField(PgDocField):
    """
    A field with a single value (e.g. String, int) calculated as an offset inside
    a (jsonb) document.
    """

    def __init__(
        self,
        name: str,
        description: str,
        alchemy_column,
        indexed: bool,
        offset: Offset | Sequence[Offset] | None = None,
        selection: str = "first",
    ) -> None:
        super().__init__(name, description, alchemy_column, indexed)
        self.offset = offset
        if selection not in SELECTION_TYPES:
            raise ValueError(
                f"Unknown field selection type {selection}. Expected one of: {(SELECTION_TYPES,)!r}"
            )
        self.aggregation = SELECTION_TYPES[selection]

    @property
    @override
    def alchemy_expression(self):
        return self._alchemy_offset_value(self.offset, self.aggregation.pg_calc).label(
            self.name
        )

    @override
    def __eq__(self, value) -> Expression:  # type: ignore[override]
        return EqualsExpression(self, value)

    @override
    def between(self, low, high) -> Expression:
        raise NotImplementedError("Simple field between expression")

    can_extract = True

    @override
    def extract(self, doc):
        return self._extract_offset_value(doc, self.offset, self.aggregation.calc)

    def evaluate(self, ctx):
        return self.extract(ctx)


class UnindexableValue(Exception):  # noqa: N818
    pass


class NumericDocField(SimpleDocField):
    type_name = "numeric"

    @override
    def value_to_alchemy(self, value):
        return cast(value, postgres.NUMERIC)

    @override
    def search_value_to_alchemy(self, value):
        if isinstance(value, float) and math.isnan(value):
            raise UnindexableValue("Cannot index NaNs")
        alc_val = self.value_to_alchemy(value)
        return func.numrange(
            alc_val,
            alc_val,
            # Inclusive on both sides.
            "[]",
            type_=NUMRANGE,
        )

    @override
    def between(self, low, high):
        # Numeric fields actually stored as ranges in current schema.
        # return ValueBetweenExpression(self, low, high)
        return RangeBetweenExpression(self, low, high, _range_class=PgRange)

    @override
    def parse_value(self, value):
        return Decimal(value)


class IntDocField(NumericDocField):
    type_name = "integer"

    @override
    def parse_value(self, value) -> int:
        return int(value)


class DoubleDocField(NumericDocField):
    type_name = "double"

    @override
    def parse_value(self, value) -> float:
        return float(value)


class BoolDocField(SimpleDocField):
    type_name = "boolean"

    @override
    def search_value_to_alchemy(self, value):
        # Convert boolean to int (range) for indexing purposes
        return func.numrange(
            int(value),
            int(value),
            # Inclusive on both sides.
            "[]",
            type_=NUMRANGE,
        )

    @override
    def value_to_alchemy(self, value):
        return cast(value, postgres.BOOLEAN)

    @override
    def parse_value(self, value) -> bool:
        if value.lower() == "false":
            return False
        if value.lower() == "true":
            return True
        return bool(value)

    @override
    def __eq__(self, value) -> Expression:  # type: ignore[override]
        # For search field comparisons???
        return EqualsExpression(self, self.search_value_to_alchemy(value))


DateFieldLike: TypeAlias = datetime | date | str | ColumnElement


class DateDocField(SimpleDocField):
    type_name = "datetime"

    @override
    def value_to_alchemy(self, value: DateFieldLike) -> DateFieldLike:
        """
        Wrap a value as needed for this field type.
        """
        if isinstance(value, datetime | str):
            return self.normalise_value(value)
        elif isinstance(value, ColumnElement):
            # SQLAlchemy expression or string are parsed in pg as dates.
            # NB: Do not cast here - casting here breaks expected behaviour in other timezones
            return value
        else:
            raise ValueError(f"Value not readable as date: {value!r}")

    @override
    def normalise_value(self, value):
        if isinstance(value, str):
            value = datetime.fromisoformat(value)
        return tz_as_utc(value)

    @override
    def search_value_to_alchemy(self, value):
        if isinstance(value, datetime):
            value = tz_as_utc(value)
        value = cast(value, TIMESTAMP(timezone=True))
        return func.tstzrange(
            value,
            value,
            # Inclusive on both sides.
            "[]",
            type_=TSTZRANGE,
        )

    @override
    def between(self, low, high):
        return ValueBetweenExpression(self, low, high)

    @override
    def parse_value(self, value) -> datetime:
        return utils.parse_time(value)

    @property
    def day(self):
        """Get field truncated to the day"""
        return NativeField(
            f"{self.name}_day",
            f"Day of {self.description}",
            self.alchemy_column,
            alchemy_expression=cast(
                func.date_trunc("day", self.alchemy_expression),
                TIMESTAMP(timezone=True),
            ),
        )


class RangeDocField(PgDocField):
    """
    A range of values. Has min and max values, which may be calculated from multiple
    values in the document.
    """

    FIELD_CLASS = SimpleDocField

    def __init__(
        self,
        name: str,
        description: str,
        alchemy_column,
        indexed: bool,
        min_offset=None,
        max_offset=None,
    ) -> None:
        super().__init__(name, description, alchemy_column, indexed)
        self.lower = self.FIELD_CLASS(
            name + "_lower",
            description,
            alchemy_column,
            indexed=False,
            offset=min_offset,
            selection="least",
        )
        self.greater = self.FIELD_CLASS(
            name + "_greater",
            description,
            alchemy_column,
            indexed=False,
            offset=max_offset,
            selection="greatest",
        )

    @override
    def value_to_alchemy(self, value):
        raise NotImplementedError("range type")

    @property
    @override
    def alchemy_expression(self):
        return self.value_to_alchemy(
            (self.lower.alchemy_expression, self.greater.alchemy_expression)
        ).label(self.name)

    @override
    def __eq__(self, value) -> Expression:  # type: ignore[override]
        # Lower and higher are interchangeable here: they're the same type.
        casted_val = self.lower.value_to_alchemy(value)
        return RangeContainsExpression(self, casted_val)

    can_extract = True

    @override
    def extract(self, doc) -> Range | None:
        min_val = self.lower.extract(doc)
        max_val = self.greater.extract(doc)
        if not min_val and not max_val:
            return None
        return Range(min_val, max_val)


class NumericRangeDocField(RangeDocField):
    FIELD_CLASS = NumericDocField
    type_name = "numeric-range"

    @override
    def value_to_alchemy(self, value):
        low, high = value
        return func.numrange(
            low,
            high,
            # Inclusive on both sides.
            "[]",
            type_=NUMRANGE,
        )

    @override
    def search_value_to_alchemy(self, value):
        low, high = value
        if isinstance(low, float) and math.isnan(low):
            raise UnindexableValue("Cannot index NaNs")
        if isinstance(high, float) and math.isnan(high):
            raise UnindexableValue("Cannot index NaNs")
        return self.value_to_alchemy(value)

    @override
    def between(self, low, high) -> Expression:
        return RangeBetweenExpression(self, low, high, _range_class=PgRange)


class IntRangeDocField(NumericRangeDocField):
    FIELD_CLASS = IntDocField
    type_name = "integer-range"


class DoubleRangeDocField(NumericRangeDocField):
    FIELD_CLASS = DoubleDocField
    type_name = "double-range"


class DateRangeDocField(RangeDocField):
    FIELD_CLASS = DateDocField
    type_name = "datetime-range"

    @override
    def value_to_alchemy(self, value):
        low, high = value
        # Is OK to cast, because we are wrapping it in timezone-aware datatype.
        if isinstance(low, ColumnElement | str):
            low = cast(low, TIMESTAMP(timezone=True))
        if isinstance(high, ColumnElement | str):
            high = cast(high, TIMESTAMP(timezone=True))
        return func.tstzrange(
            low,
            high,
            # Inclusive on both sides.
            "[]",
            type_=TSTZRANGE,
        )

    @override
    def search_value_to_alchemy(self, value):
        low, high = value
        if isinstance(low, datetime):
            low = tz_as_utc(low)
        if isinstance(high, datetime):
            high = tz_as_utc(high)
        return func.tstzrange(
            low,
            high,
            # Inclusive on both sides.
            "[]",
            type_=TSTZRANGE,
        )

    @override
    def normalise_value(self, value):
        if isinstance(value, datetime):
            return tz_as_utc(value)
        elif isinstance(value, PgRange):
            return PgRange(
                lower=tz_as_utc(value.lower),
                upper=tz_as_utc(value.upper),
                bounds=value.bounds,
            )
        else:
            return tuple(tz_as_utc(v) for v in value)

    @override
    def between(self, low, high) -> Expression:
        low = _number_implies_year(low)
        high = _number_implies_year(high)

        if isinstance(low, datetime) and isinstance(high, datetime):
            return RangeBetweenExpression(
                self,
                tz_as_utc(low).astimezone(timezone.utc),
                tz_as_utc(high).astimezone(timezone.utc),
                _range_class=PgRange,
            )
        else:
            raise ValueError(
                "Unknown comparison type for date range: "
                f"expecting datetimes, got: ({low!r}, {high!r})"
            )

    @property
    def expression_with_leniency(self):
        low = cast(self.lower.alchemy_expression, TIMESTAMP(timezone=True)) - cast(
            "500 milliseconds", INTERVAL
        )
        high = cast(self.greater.alchemy_expression, TIMESTAMP(timezone=True)) + cast(
            "500 milliseconds", INTERVAL
        )
        return func.tstzrange(
            low,
            high,
            # Inclusive on both sides.
            "[]",
            type_=TSTZRANGE,
        )


def _number_implies_year(v: int | float | datetime) -> datetime:
    """
    >>> _number_implies_year(1994)
    datetime.datetime(1994, 1, 1, 0, 0)
    >>> _number_implies_year(datetime(1994, 4, 4))
    datetime.datetime(1994, 4, 4, 0, 0)
    """
    if isinstance(v, int):
        return datetime(v, 1, 1)
    # The expression module parses all number ranges as floats.
    if isinstance(v, float):
        return datetime(int(v), 1, 1)

    return v


class PgExpression(Expression):
    def __init__(self, field: PgField) -> None:
        super().__init__()
        self.field = field

    @property
    def alchemy_expression(self):
        """
        Get an SQLAlchemy expression for accessing this field.
        """
        raise NotImplementedError("alchemy expression")


class ValueBetweenExpression(PgExpression):
    def __init__(self, field, low_value, high_value) -> None:
        super().__init__(field)
        self.low_value = low_value
        self.high_value = high_value

    @property
    @override
    def alchemy_expression(self):
        if self.low_value is not None and self.high_value is not None:
            return and_(
                self.field.search_alchemy_expression >= self.low_value,
                self.field.search_alchemy_expression <= self.high_value,
            )
        if self.low_value is not None:
            return self.field.search_alchemy_expression >= self.low_value
        if self.high_value is not None:
            return self.field.search_alchemy_expression <= self.high_value
        raise ValueError("Expect at least one of [low,high] to be set")


class RangeBetweenExpression(PgExpression):
    def __init__(self, field, low_value, high_value, _range_class) -> None:
        super().__init__(field)
        self.low_value = low_value
        self.high_value = high_value
        self._range_class = _range_class
        self._alc_val = self._range_class(self.low_value, self.high_value, bounds="[]")

    @property
    @override
    def alchemy_expression(self):
        return self.field.search_alchemy_expression.overlaps(self._alc_val)


class RangeContainsExpression(PgExpression):
    def __init__(self, field, value) -> None:
        super().__init__(field)
        self.value = value

    @property
    @override
    def alchemy_expression(self):
        return self.field.search_alchemy_expression.contains(self.value)


class EqualsExpression(PgExpression):
    def __init__(self, field, value) -> None:
        super().__init__(field)
        self.value = value

    @property
    @override
    def alchemy_expression(self):
        return self.field.search_alchemy_expression == self.value

    @override
    def evaluate(self, ctx):
        return self.field.evaluate(ctx) == self.value


def parse_fields(doc: dict, table_column) -> dict[str, PgField]:
    """
    Parse a field spec document into objects.

    Example document:

    ::

        {
            # Field name:
            'lat': {
                # Field type & properties.
                'type': 'float-range',
                'min_offset': [
                    # Offsets within a dataset document for this field.
                    ['extent', 'coord', 'ul', 'lat'],
                    ['extent', 'coord', 'll', 'lat']
                ],
                'max_offset': [
                    ['extent', 'coord', 'ur', 'lat'],
                    ['extent', 'coord', 'lr', 'lat']
                ]
            }
        }

    :param table_column: SQLAlchemy jsonb column for the document we're reading fields from.
    """
    # Implementations of fields for this driver
    types = {
        SimpleDocField,
        IntDocField,
        BoolDocField,
        DoubleDocField,
        DateDocField,
        NumericRangeDocField,
        IntRangeDocField,
        DoubleRangeDocField,
        DateRangeDocField,
    }
    type_map = {f.type_name: f for f in types}
    # An alias for backwards compatibility
    type_map["float-range"] = NumericRangeDocField

    # No later field should have overridden string
    assert type_map["string"] == SimpleDocField

    def _get_field(name: str, descriptor: dict, column) -> PgField:
        """
        :param column: SQLAlchemy table column
        """
        ctorargs = descriptor.copy()
        type_name = ctorargs.pop("type", "string")
        description = ctorargs.pop("description", None)
        indexed_val = ctorargs.pop("indexed", "true")
        indexed = (
            indexed_val.lower() == "true"
            if isinstance(indexed_val, str)
            else indexed_val
        )

        field_class = type_map.get(type_name)
        if not field_class:
            raise ValueError(
                f"Field {name!r} has unknown type {type_name!r}."
                f" Available types are: {list(type_map.keys())!r}"
            )
        try:
            return field_class(name, description, column, indexed, **ctorargs)
        except TypeError as e:
            raise RuntimeError(
                f"Field {name} has unexpected argument for a {type_name}", e
            ) from None

    return {
        name: _get_field(name, descriptor, table_column)
        for name, descriptor in doc.items()
    }


def _coalesce(*values):
    """
    Return first non-none value.
    Return None if all values are None, or there are no values passed in.

    >>> _coalesce(1, 2)
    1
    >>> _coalesce(None, 2, 3)
    2
    >>> _coalesce(None, None, 3, None, 5)
    3
    """
    for v in values:
        if v is not None:
            return v
    return None


# How to choose/combine multiple doc values.
ValueAggregation = namedtuple("ValueAggregation", ("calc", "pg_calc"))
SELECTION_TYPES = {
    # First non-null
    "first": ValueAggregation(_coalesce, func.coalesce),
    # min/max
    "least": ValueAggregation(min, func.least),
    "greatest": ValueAggregation(max, func.greatest),
}



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

import datetime
from collections.abc import Generator, Iterable, Mapping, Sequence
from typing import Any, NamedTuple

from deprecat import deprecat
from typing_extensions import override

from datacube.index.abstract import DSID, AbstractDatasetResource
from datacube.migration import ODC2DeprecationWarning
from datacube.model import Dataset, Product, QueryDict, QueryField
from datacube.utils.changes import Change, Offset


class DatasetResource(AbstractDatasetResource):
    def __init__(self, index) -> None:
        super().__init__(index)

    @override
    def get_unsafe(
        self,
        id_: DSID,
        include_sources: bool = False,
        include_deriveds: bool = False,
        max_depth: int = 0,
    ):
        raise KeyError(id_)

    @override
    def bulk_get(self, ids) -> list:
        return []

    @override
    def get_derived(self, id_) -> list:
        return []

    @override
    def has(self, id_) -> bool:
        return False

    @override
    def bulk_has(self, ids_) -> list:
        return [False for id_ in ids_]

    @override
    def add(
        self,
        dataset: Dataset,
        with_lineage: bool = True,
        archive_less_mature: int | None = None,
    ) -> Dataset:
        raise NotImplementedError()

    @override
    def search_product_duplicates(self, product: Product, *args) -> list:
        return []

    @override
    def can_update(
        self, dataset, updates_allowed=None
    ) -> tuple[bool, list[Change], list[Change]]:
        raise NotImplementedError()

    @override
    def update(self, dataset: Dataset, updates_allowed=None, archive_less_mature=None):
        raise NotImplementedError()

    @override
    def archive(self, ids):
        raise NotImplementedError()

    @override
    def restore(self, ids):
        raise NotImplementedError()

    @override
    def purge(self, ids: Iterable[DSID], allow_delete_active: bool = False) -> Sequence:
        raise NotImplementedError()

    @override
    def get_all_dataset_ids(self, archived: bool) -> list:
        return []

    @deprecat(
        reason="Multiple locations per dataset are now deprecated.  Please use the 'get_location' method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def get_locations(self, id_) -> list:
        return []

    @override
    def get_location(self, id_) -> None:
        return None

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Archived locations may not be accessible in future releases.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def get_archived_locations(self, id_) -> list:
        return []

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Archived locations may not be accessible in future releases.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def get_archived_location_times(self, id_) -> list:
        return []

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def add_location(self, id_, uri) -> bool:
        raise NotImplementedError()

    @override
    def get_datasets_for_location(self, uri, mode=None) -> list:
        return []

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def remove_location(self, id_, uri) -> bool:
        raise NotImplementedError()

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Archived locations may not be accessible in future releases. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def archive_location(self, id_, uri) -> bool:
        raise NotImplementedError()

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Archived locations may not be restorable in future releases. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def restore_location(self, id_, uri) -> bool:
        raise NotImplementedError()

    @override
    def search_by_metadata(self, metadata, archived: bool | None = False) -> list:
        return []

    @deprecat(
        deprecated_args={
            "source_filter": {
                "reason": "Filtering by source metadata is deprecated and will be removed in future.",
                "version": "1.9.0",
                "category": ODC2DeprecationWarning,
            }
        }
    )
    @override
    def search(
        self,
        limit: int | None = None,
        source_filter: QueryDict | None = None,
        archived: bool | None = False,
        order_by: Iterable[Any] | None = None,
        **query: QueryField,
    ) -> Iterable[Dataset]:
        return []

    @override
    def search_by_product(
        self, archived: bool | None = False, **query: QueryField
    ) -> Iterable[tuple[Product, Iterable[Dataset]]]:
        return []

    @override
    def search_returning(
        self,
        field_names: Iterable[str] | None = None,
        custom_offsets: Mapping[str, Offset] | None = None,
        limit: int | None = None,
        archived: bool | None = False,
        order_by: Iterable[Any] | None = None,
        **query: QueryField,
    ) -> Generator[NamedTuple]:
        yield from []

    @override
    def count(self, archived: bool | None = False, **query) -> int:
        return 0

    @override
    def count_by_product(self, archived: bool | None = False, **query) -> list:
        return []

    @override
    def count_by_product_through_time(
        self, period: str, archived: bool | None = False, **query
    ) -> list:
        return []

    @override
    def count_product_through_time(
        self, period: str, archived: bool | None = False, **query
    ) -> list:
        return []

    @deprecat(
        reason="This method is deprecated and will be removed in 2.0.  "
        "Consider migrating to search_returning()",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @override
    def search_summaries(self, **query) -> list:
        return []

    @override
    def temporal_extent(
        self, ids: Iterable[DSID]
    ) -> tuple[datetime.datetime, datetime.datetime]:
        raise KeyError(str(ids))

    # pylint: disable=redefined-outer-name
    @override
    def search_returning_datasets_light(
        self,
        field_names: tuple,
        custom_offsets=None,
        limit: int | None = None,
        archived: bool | None = False,
        **query,
    ) -> list:
        return []

    @override
    def spatial_extent(self, ids=None, product=None, crs=None):
        return None
